# -*- coding: utf-8 -*-
"""
router handlers base module.
"""

import time

from threading import Lock

from werkzeug.routing import Rule

import pyrin.processor.response.services as response_services
import pyrin.processor.response.status.services as status_services
import pyrin.configuration.services as config_services
import pyrin.security.session.services as session_services
import pyrin.utils.misc as misc_utils
import pyrin.utils.headers as header_utils
import pyrin.utils.function as func_utils

from pyrin.core.globals import _
from pyrin.processor.cors.structs import CORS
from pyrin.api.schema.structs import ResultSchema
from pyrin.core.enumerations import HTTPMethodEnum
from pyrin.database.paging.paginator import SimplePaginator
from pyrin.processor.response.wrappers.base import CoreResponse
from pyrin.processor.response.enumerations import ResponseHeaderEnum
from pyrin.api.router.handlers.exceptions import InvalidViewFunctionTypeError, \
    MaxContentLengthLimitMismatchError, InvalidResultSchemaTypeError, \
    RouteIsNotBoundedToMapError, RouteIsNotBoundedError, InvalidResponseStatusCodeError, \
    RequestLimitOrLifetimeRequiredError, InvalidRequestLimitError, \
    InvalidLifetimeError, URLNotFoundError, ViewFunctionRequiredParamsError


class RouteBase(Rule):
    """
    route base class.

    this class should be used as base for all route types.
    """

    # these are http methods that will not be considered as operational.
    NON_OPERATIONAL_METHODS = [HTTPMethodEnum.HEAD,
                               HTTPMethodEnum.OPTIONS]

    result_schema_class = ResultSchema
    paginator_class = SimplePaginator
    cors_class = CORS

    def __init__(self, rule, **options):
        """
        initializes an instance of RouteBase.

        :param str rule: unique url rule to register this route for.
                         routes with duplicated urls and http methods will be
                         overwritten if `replace=True` option is provided.
                         otherwise an error will be raised.

        :keyword str | tuple[str] methods: http methods that this route could handle.
                                           if not provided, defaults to `GET`, `HEAD`
                                           and `OPTIONS`.
                        
        :keyword function view_function: a function to be called on accessing this route.

        :keyword str endpoint: the endpoint for the registered url rule.

        :keyword dict defaults: an optional dict with defaults for other rules with the
                                same endpoint. this is a bit tricky but useful if you
                                want to have unique urls.

        :keyword str subdomain: the subdomain rule string for this rule. If not specified the
                                rule only matches for the `default_subdomain` of the map. if
                                the map is not bound to a subdomain this feature is disabled.

        :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only for
                                      this rule. if not specified the `Map` setting is used.

        :keyword bool merge_slashes: override `Map.merge_slashes` for this rule.

        :keyword bool build_only: set this to True and the rule will never match but will
                                  create a url that can be build. this is useful if you have
                                  resources on a subdomain or folder that are not handled by
                                  the WSGI application (like static data)

        :keyword str | callable redirect_to: if given this must be either a string
                                             or callable. in case of a callable it's
                                             called with the url adapter that
                                             triggered the match and the values
                                             of the url as keyword arguments and has
                                             to return the target for the redirect,
                                             otherwise it has to be a string with
                                             placeholders in rule syntax.

        :keyword bool alias: if enabled this rule serves as an alias for another rule with
                             the same endpoint and arguments.

        :keyword str host: if provided and the url map has host matching enabled this can be
                           used to provide a match rule for the whole host. this also means
                           that the subdomain feature is disabled.

        :keyword bool websocket: if set to True, this rule is only matches for
                                 websocket (`ws://`, `wss://`) requests. by default,
                                 rules will only match for http requests.
                                 defaults to False if not provided.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :keyword int status_code: status code to be returned on successful responses.
                                  defaults to corresponding status code of request's
                                  http method if not provided.

        :note status_code: it could be a value from `InformationResponseCodeEnum`
                           or `SuccessfulResponseCodeEnum` or `RedirectionResponseCodeEnum`.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as processed if it is from
                                     `InformationResponseCodeEnum` or
                                     `SuccessfulResponseCodeEnum` or
                                     `RedirectionResponseCodeEnum` values. otherwise
                                     all codes from `INFORMATION_CODE_MIN`
                                     to `INFORMATION_CODE_MAX` or from
                                     `SUCCESS_CODE_MIN` to `SUCCESS_CODE_MAX`
                                     or from `REDIRECTION_CODE_MIN` to
                                     `REDIRECTION_CODE_MAX` will be considered
                                     as processed. defaults to True if not provided.

        :keyword ResultSchema | type[ResultSchema] result_schema: result schema to be used
                                                                  to filter results. it could
                                                                  be an instance or a type
                                                                  of `ResultSchema` class.

        :keyword bool indexed: specifies that list results must
                               include an extra field as row index.
                               the name of the index field and the initial value
                               of index could be provided by `index_name` and
                               `start_index` respectively. `indexed` keyword has
                               only effect if the returning result contains a list
                               of objects.

        :keyword str index_name: name of the extra field to contain
                                 the row index of each result. if not provided
                                 defaults to `row_num` value.

        :keyword int start_index: the initial value of row index. if not
                                  provided, starts from 1.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults
                                                      to `SECURE_TRUE` if not provided. it
                                                      will be used only for entity conversion.
                                                      this value will override the
                                                      corresponding value of `result_schema`
                                                      if provided.

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.
                            it will be used only for entity conversion.
                            this value will override the corresponding value of
                            `result_schema` if provided.

        :keyword bool no_cache: a value indicating that the response returning from this route
                                must have a `Cache-Control: no-cache` header. this header will
                                be automatically added. defaults to False if not provided.

        :keyword bool paged: specifies that this route should return paginated results.
                             defaults to False if not provided.

        :keyword int page_size: default page size for this route.
                                defaults to `default_page_size` from
                                `database` config store if not provided.

        :keyword int max_page_size: maximum page size that client is allowed
                                    to request for this route. defaults to
                                    `max_page_size` from `database` configs store
                                    if not provided.

        :keyword bool cors_enabled: specifies that cross origin resource sharing is enabled.
                                    if not provided, it will be get from cors config store.

        :keyword bool cors_always_send: specifies that cors headers must be included in
                                        response even if the request does not have origin header.
                                        if not provided, it will be get from cors config store.

        :keyword list[str] cors_allowed_origins: a list of extra allowed origins to be used
                                                 in conjunction with default allowed ones.

        :keyword list[str] cors_exposed_headers: extra exposed headers to be combined
                                                 with default ones.

        :keyword list[str] cors_allowed_headers: extra allowed headers to be combined
                                                 with default ones.

        :keyword bool cors_allow_credentials: specifies that browsers are allowed to pass
                                              response headers to front-end javascript code
                                              if the route is authenticated.
                                              if not provided, it will be get from cors config
                                              store.

        :keyword int cors_max_age: maximum number of seconds to cache results.
                                   if not provided, it will be get from cors config store.

        :raises PageSizeLimitError: page size limit error.
        :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
        :raises InvalidViewFunctionTypeError: invalid view function type error.
        :raises InvalidResultSchemaTypeError: invalid result schema type error.
        :raises InvalidResponseStatusCodeError: invalid response status code error.
        """

        methods = options.get('methods', None)
        methods = misc_utils.make_iterable(methods, tuple)
        options.update(methods=methods)

        self._namespace = config_services.get('api', 'general', 'namespace', default=None)
        if self._namespace not in (None, '') and not self._namespace.isspace():
            rule = '{namespace}{rule}'.format(namespace=self._namespace, rule=rule)

        append_slash = config_services.get('api', 'general', 'append_slash', default=True)
        if not rule.endswith('/') and append_slash is True:
            rule = '{rule}/'.format(rule=rule)

        # we should call super method with exact param names because it
        # does not have `**options` in it's signature and raises an error
        # if extra keywords passed to it.
        super().__init__(rule,
                         defaults=options.get('defaults', None),
                         subdomain=options.get('subdomain', None),
                         methods=methods,
                         build_only=options.get('build_only', False),
                         endpoint=options.get('endpoint', None),
                         strict_slashes=options.get('strict_slashes', None),
                         merge_slashes=options.get('merge_slashes', None),
                         redirect_to=options.get('redirect_to', None),
                         alias=options.get('alias', False),
                         host=options.get('host', None),
                         websocket=options.get('websocket', False))

        self._view_function = options.get('view_function')
        self._paginator = self._get_paginator(**options)
        self._result_schema = self._extract_schema(**options)
        self._cors = self._get_cors_configs(**options)

        global_limit = config_services.get('api', 'general', 'max_content_length')
        restricted_length = options.get('max_content_length',
                                        config_services.get('api',
                                                            'general',
                                                            'restricted_max_content_length'))
        if restricted_length > global_limit:
            raise MaxContentLengthLimitMismatchError('Specified max content length '
                                                     '[{restricted}] for route [{route}] is '
                                                     'higher than global limit which is '
                                                     '[{global_limit}].'
                                                     .format(restricted=restricted_length,
                                                             route=self.rule,
                                                             global_limit=global_limit))

        self._max_content_length = restricted_length

        full_name = misc_utils.try_get_fully_qualified_name(self._view_function)
        if not callable(self._view_function):
            raise InvalidViewFunctionTypeError('The provided view function [{function}] '
                                               'for route with url [{url}] is not callable.'
                                               .format(function=full_name,
                                                       url=self.rule))

        self._required_arguments = func_utils.get_required_arguments(self._view_function)
        self._no_cache = options.get('no_cache', False)

        status_code = options.pop('status_code', None)
        if status_code is not None and \
                status_services.is_processed(status_code, **options) is not True:
            raise InvalidResponseStatusCodeError('The provided status code [{status}] for '
                                                 'route [{route}] on view function [{function}] '
                                                 'must be from information, success or '
                                                 'redirection codes. and if you want to return '
                                                 'a status code for errors, you should raise '
                                                 'an exception with relevant code as status '
                                                 'code inside your method. pyrin will handle '
                                                 'exceptions and converts them to correct '
                                                 'responses.'
                                                 .format(status=status_code,
                                                         route=self.rule,
                                                         function=full_name))
        self._status_code = status_code

    def __eq__(self, other):
        """
        gets a value indicating that current route is equal to provided route.

        note that for two routes to be considered equal, three conditions must be met:

        1. both routes be an instance of `RouteBase`.
        2. both routes have the same exact url rule.
        3. both routes have the same exact http methods.

        this has some differences with how flask compares routes.
        and this is required because pyrin handles routes on its own and assures
        that there should not be multiple routes with the same url and http methods.

        :param object other: other instance to be compared to current route.

        :rtype: bool
        """

        if not isinstance(other, RouteBase):
            return False

        return self._trace == other._trace and \
            sorted(set(self.methods)) == sorted(set(other.methods))

    def unbind(self, map):
        """
        sets this route's map to None.

        :param Map map: the map that this route is bounded to.
                        it must be the exact map of this route.

        :raises RouteIsNotBoundedError: route is not bounded error.
        :raises RouteIsNotBoundedToMapError: route is not bounded to map error.
        """

        if self.map is None:
            raise RouteIsNotBoundedError('Route [{route}] is not bounded to any Map.'
                                         .format(route=self))

        if self.map is not map:
            raise RouteIsNotBoundedToMapError('Route [{route}] is not bounded '
                                              'to provided Map [{map}].'
                                              .format(route=self, map=map))

        self.map = None

    def get_duplicate_methods(self, methods):
        """
        gets a tuple of this route's methods that are duplicated with provided methods.

        it gets an empty tuple if no duplication found.
        note that any of `NON_OPERATIONAL_METHODS` will not be considered
        for duplication. because these two methods will be handled correctly by
        flask itself. but if you want to explicitly define routes for these type
        of methods, you have to implement the required parts manually.

        :param str | tuple[str] methods: http methods to check for duplication.

        :returns: tuple of duplicate methods.
        :rtype: tuple[str]
        """

        methods = misc_utils.make_iterable(methods, set)
        duplicate_methods = set(self.methods).intersection(methods)
        duplicate_methods = duplicate_methods.difference(set(self.NON_OPERATIONAL_METHODS))

        return tuple(duplicate_methods)

    def remove_methods(self, methods):
        """
        removes the provided methods from this route's methods.

        note that this will not remove any of `NON_OPERATIONAL_METHODS`.

        :param str | tuple[str] methods: http methods to be removed.
        """

        to_be_removed = self.get_duplicate_methods(methods)
        updated_methods = set(self.methods).difference(set(to_be_removed))
        self.methods = updated_methods
        self.refresh()

    def handle(self, inputs, **options):
        """
        handles the current route.

        :param dict inputs: view function inputs.

        :raises ViewFunctionRequiredParamsError: view function required params error.

        :returns: view function's result.
        :rtype: object
        """

        if self._result_schema is not None:
            self._inject_result_schema()

        if self._paginator is not None:
            self._inject_paginator(inputs, **options)

        self._handle(inputs, **options)
        result = self._call_view_function(inputs, **options)
        self._finished(result, **options)

        return self._prepare_response(result)

    def _get_paginator(self, **options):
        """
        gets paginator for this route if required.

        :keyword bool paged: specifies that this route should return paginated results.
                             defaults to False if not provided.

        :keyword int page_size: default page size for this route.
                                defaults to `default_page_size` from
                                `database` config store if not provided.

        :keyword int max_page_size: maximum page size that client is allowed
                                    to request for this route. defaults to
                                    `max_page_size` from `database` configs store
                                    if not provided.

        :raises PageSizeLimitError: page size limit error.

        :rtype: PaginatorBase
        """

        paged = options.get('paged', False)
        if paged is True:
            return self.paginator_class(**options)

        return None

    def _get_cors_configs(self, **options):
        """
        gets cors configs for this route if enabled.

        :keyword bool cors_enabled: specifies that cross origin resource sharing is enabled.
                                    defaults to False if not provided.

        :keyword bool cors_always_send: specifies that cors headers must be included in
                                        response even if the request does not have origin header.
                                        if not provided, it will be get from cors config store.

        :keyword list[str] cors_allowed_origins: a list of extra allowed origins to be used
                                                 in conjunction with default allowed ones.

        :keyword list[str] cors_exposed_headers: extra exposed headers to be combined
                                                 with default ones.

        :keyword list[str] cors_allowed_headers: extra allowed headers to be combined
                                                 with default ones.

        :keyword bool cors_allow_credentials: specifies that browsers are allowed to pass
                                              response headers to front-end javascript code
                                              if the route is authenticated.
                                              if not provided, it will be get from cors config
                                              store.

        :keyword int cors_max_age: maximum number of seconds to cache results.
                                   if not provided, it will be get from cors config store.
        """

        return self.cors_class(**options)

    def _handle(self, inputs, **options):
        """
        handles the current route.

        routes which need to perform extra operations before
        view function execution, must override this method.

        :param dict inputs: view function inputs.
        """
        pass

    def _finished(self, result, **options):
        """
        does the extra required works after the view function has been executed.

        routes which need to perform extra operations after
        view function execution, must override this method.

        :param object result: view function execution result.
        """
        pass

    def _extract_schema(self, **options):
        """
        extracts schema related attributes from given values and returns an schema object.

        if no schema related item is provided, it returns None.

        :keyword ResultSchema | type[ResultSchema] result_schema: result schema to be used
                                                                  to filter results. it could
                                                                  be an instance or a type
                                                                  of `ResultSchema` class.

        :keyword bool indexed: specifies that list results must
                               include an extra field as row index.
                               the name of the index field and the initial value
                               of index could be provided by `index_name` and
                               `start_index` respectively. `indexed` keyword has
                               only effect if the returning result contains a list
                               of objects.

        :keyword str index_name: name of the extra field to contain
                                 the row index of each result. if not provided
                                 defaults to `row_num` value.

        :keyword int start_index: the initial value of row index. if not
                                  provided, starts from 1.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults to
                                                      `SECURE_TRUE` if not provided. it will
                                                      be used only for entity conversion.

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.
                            it will be used only for entity conversion.
                            this value will override the corresponding value of
                            `result_schema` if provided.

        :raises InvalidResultSchemaTypeError: invalid result schema type error.

        :rtype: ResultSchema
        """

        result_schema = options.get('result_schema')
        is_subclass = isinstance(result_schema, type) and issubclass(result_schema, ResultSchema)
        is_instance = isinstance(result_schema, ResultSchema)
        if result_schema is not None and not is_instance and not is_subclass:
            raise InvalidResultSchemaTypeError('Input parameter [{instance}] is not '
                                               'an instance or subclass of [{base}].'
                                               .format(instance=result_schema,
                                                       base=ResultSchema))

        readable = options.get('readable')
        depth = options.get('depth')
        indexed = options.get('indexed')
        index_name = options.get('index_name')
        start_index = options.get('start_index')

        if result_schema is not None and is_subclass:
            result_schema = result_schema()

        if result_schema is None and (readable is not None or
                                      depth is not None or indexed is True):
            return self.result_schema_class(readable=readable, depth=depth,
                                            indexed=indexed, index_name=index_name,
                                            start_index=start_index)

        elif result_schema is not None and (readable is not None or depth is not None or
                                            indexed is not None or index_name is not None or
                                            start_index is not None):
            updated_schema = result_schema.copy()
            if readable is not None:
                updated_schema.readable = readable
            if depth is not None:
                updated_schema.depth = depth
            if indexed is not None:
                updated_schema.indexed = indexed
            if index_name is not None:
                updated_schema.index_name = index_name
            if start_index is not None:
                updated_schema.start_index = start_index

            return updated_schema

        return result_schema

    def _prepare_response(self, response):
        """
        prepares given response to be returned to client.

        it does some operations such as adding this route's status code
        into response. if the response from view function already is a
        tuple with status code, then the status code of this route
        won't be injected.

        :param object | tuple | dict | CoreResponse response: response value to be prepared.

        :note response: it could be an object or a dict or a tuple with the length of 2 or 3.
                        in the form of: body
                                        body, status_code
                                        body, headers
                                        body, status_code, header

        :rtype: tuple[object, int, dict | Headers] | object
        """

        body, status_code, headers = response_services.unpack_response(response)
        if status_code is None and self.status_code is not None:
            status_code = self.status_code
        elif status_code is None and self.status_code is None:
            status_code = status_services.get_status_code()

        headers = self._prepare_headers(headers)
        return response_services.pack_response(body, status_code, headers)

    def _prepare_headers(self, headers):
        """
        prepares required headers into given dict.

        :param dict | Headers headers: headers dict to be prepared.

        :rtype: CoreHeaders
        """

        if self._no_cache is True:
            headers = header_utils.convert_headers(headers, ignore_null=False)
            cache_header = {ResponseHeaderEnum.CACHE_CONTROL: 'no-cache'}
            headers.extend(cache_header)

        return headers

    def _inject_result_schema(self):
        """
        injects this route's result schema into current request context.
        """

        session_services.add_request_context('result_schema', self._result_schema)

    def _inject_paginator(self, inputs, **options):
        """
        injects this route's paginator into current request context.

        :param dict inputs: view function inputs.
        """

        paginator = self._paginator.copy()
        request = session_services.get_current_request()
        paging_params = request.get_paging_params()
        paginator.inject_paging_keys(inputs, **paging_params)
        session_services.add_request_context('paginator', paginator)

    def _call_view_function(self, inputs, **options):
        """
        calls the route's view function.

        :param dict inputs: view function inputs.

        :raises ViewFunctionRequiredParamsError: view function required params error.

        :returns: view function's result.
        :rtype: object
        """

        self._check_required_arguments(inputs)
        return self.view_function(**inputs)

    def _check_required_arguments(self, inputs):
        """
        checks that all required arguments for this route are present in inputs.

        :raises ViewFunctionRequiredParamsError: view function required params error.
        """

        not_present = self._required_arguments.difference(set(inputs.keys()))
        present = self._required_arguments.difference(not_present)
        not_present = list(not_present)

        for item in present:
            if inputs.get(item) is None:
                not_present.append(item)

        if len(not_present) > 0:
            raise ViewFunctionRequiredParamsError(_('These values are required: {params}')
                                                  .format(params=not_present))

    @property
    def view_function(self):
        """
        gets this route's view function.

        :rtype: callable
        """

        return self._view_function

    @property
    def required_arguments(self):
        """
        gets this route's required arguments.

        :rtype: set[str]
        """

        return self._required_arguments

    @property
    def max_content_length(self):
        """
        gets this route's max content length in bytes.

        :rtype: int
        """

        return self._max_content_length

    @property
    def result_schema(self):
        """
        gets this route's result schema if available.

        :rtype: ResultSchema
        """

        return self._result_schema

    @property
    def is_operational(self):
        """
        gets a value indicating that this route has any operational http methods.

        note that any of `NON_OPERATIONAL_METHODS` will not be considered as operational.

        :rtype: bool
        """

        operational_methods = set(self.methods).difference(set(self.NON_OPERATIONAL_METHODS))
        return len(operational_methods) > 0

    @property
    def status_code(self):
        """
        gets this route's success response status code.

        it could be None if not provided.

        :rtype: int
        """

        return self._status_code

    @property
    def cors(self):
        """
        gets cors attribute of this route.

        :rtype: CORS
        """

        return self._cors


class TemporaryRouteBase(RouteBase):
    """
    temporary route bass class.

    this class should be used as base for all temporary route types.
    this type of routes will unregister themselves after a number of requests
    or a period of time after registration.
    """

    def __init__(self, rule, **options):
        """
        initializes an instance of TemporaryRouteBase.

        :param str rule: unique url rule to register this route for.
                         routes with duplicated urls and http methods will be
                         overwritten if `replace=True` option is provided.
                         otherwise an error will be raised.

        :keyword int request_limit: number of allowed requests to this
                                    route before it unregisters itself.
                                    defaults to None if not Provided.

        :keyword int lifetime: number of seconds in which this route must remain
                               responsive after initial registration. after this
                               period, the route will unregister itself.
                               defaults to None if not provided.

        :note request_limit, lifetime: if both of these values are provided, this
                                       route will unregister itself if any of
                                       these two conditions are met.

        :keyword str | tuple[str] methods: http methods that this route could handle.
                                           if not provided, defaults to `GET`, `HEAD`
                                           and `OPTIONS`.

        :keyword function view_function: a function to be called on accessing this route.

        :keyword str endpoint: the endpoint for the registered url rule.

        :keyword dict defaults: an optional dict with defaults for other rules with the
                                same endpoint. this is a bit tricky but useful if you
                                want to have unique urls.

        :keyword str subdomain: the subdomain rule string for this rule. If not specified the
                                rule only matches for the `default_subdomain` of the map. if
                                the map is not bound to a subdomain this feature is disabled.

        :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only for
                                      this rule. if not specified the `Map` setting is used.

        :keyword bool merge_slashes: override `Map.merge_slashes` for this rule.

        :keyword bool build_only: set this to True and the rule will never match but will
                                  create a url that can be build. this is useful if you have
                                  resources on a subdomain or folder that are not handled by
                                  the WSGI application (like static data)

        :keyword str | callable redirect_to: if given this must be either a string
                                             or callable. in case of a callable it's
                                             called with the url adapter that
                                             triggered the match and the values
                                             of the url as keyword arguments and has
                                             to return the target for the redirect,
                                             otherwise it has to be a string with
                                             placeholders in rule syntax.

        :keyword bool alias: if enabled this rule serves as an alias for another rule with
                             the same endpoint and arguments.

        :keyword str host: if provided and the url map has host matching enabled this can be
                           used to provide a match rule for the whole host. this also means
                           that the subdomain feature is disabled.

        :keyword bool websocket: if set to True, this rule is only matches for
                                 websocket (`ws://`, `wss://`) requests. by default,
                                 rules will only match for http requests.
                                 defaults to False if not provided.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :keyword int status_code: status code to be returned on successful responses.
                                  defaults to corresponding status code of request's
                                  http method if not provided.

        :note status_code: it could be a value from `InformationResponseCodeEnum`
                           or `SuccessfulResponseCodeEnum` or `RedirectionResponseCodeEnum`.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as processed if it is from
                                     `InformationResponseCodeEnum` or
                                     `SuccessfulResponseCodeEnum` or
                                     `RedirectionResponseCodeEnum` values. otherwise
                                     all codes from `INFORMATION_CODE_MIN`
                                     to `INFORMATION_CODE_MAX` or from
                                     `SUCCESS_CODE_MIN` to `SUCCESS_CODE_MAX`
                                     or from `REDIRECTION_CODE_MIN` to
                                     `REDIRECTION_CODE_MAX` will be considered
                                     as processed. defaults to True if not provided.

        :keyword ResultSchema | type[ResultSchema] result_schema: result schema to be used
                                                                  to filter results. it could
                                                                  be an instance or a type
                                                                  of `ResultSchema` class.

        :keyword bool indexed: specifies that list results must
                               include an extra field as row index.
                               the name of the index field and the initial value
                               of index could be provided by `index_name` and
                               `start_index` respectively. `indexed` keyword has
                               only effect if the returning result contains a list
                               of objects.

        :keyword str index_name: name of the extra field to contain
                                 the row index of each result. if not provided
                                 defaults to `row_num` value.

        :keyword int start_index: the initial value of row index. if not
                                  provided, starts from 1.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults
                                                      to `SECURE_TRUE` if not provided. it
                                                      will be used only for entity conversion.
                                                      this value will override the
                                                      corresponding value of `result_schema`
                                                      if provided.

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.
                            it will be used only for entity conversion.
                            this value will override the corresponding value of
                            `result_schema` if provided.

        :keyword bool no_cache: a value indicating that the response returning from this route
                                must have a `Cache-Control: no-cache` header. this header will
                                be automatically added. defaults to False if not provided.

        :keyword bool paged: specifies that this route should return paginated results.
                             defaults to False if not provided.

        :keyword int page_size: default page size for this route.
                                defaults to `default_page_size` from
                                `database` config store if not provided.

        :keyword int max_page_size: maximum page size that client is allowed
                                    to request for this route. defaults to
                                    `max_page_size` from `database` configs store
                                    if not provided.

        :keyword bool cors_enabled: specifies that cross origin resource sharing is enabled.
                                    if not provided, it will be get from cors config store.

        :keyword bool cors_always_send: specifies that cors headers must be included in
                                        response even if the request does not have origin header.
                                        if not provided, it will be get from cors config store.

        :keyword list[str] cors_allowed_origins: a list of extra allowed origins to be used
                                                 in conjunction with default allowed ones.

        :keyword list[str] cors_exposed_headers: extra exposed headers to be combined
                                                 with default ones.

        :keyword list[str] cors_allowed_headers: extra allowed headers to be combined
                                                 with default ones.

        :keyword bool cors_allow_credentials: specifies that browsers are allowed to pass
                                              response headers to front-end javascript code
                                              if the route is authenticated.
                                              if not provided, it will be get from cors config
                                              store.

        :keyword int cors_max_age: maximum number of seconds to cache results.
                                   if not provided, it will be get from cors config store.

        :raises PageSizeLimitError: page size limit error.
        :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
        :raises InvalidViewFunctionTypeError: invalid view function type error.
        :raises InvalidResultSchemaTypeError: invalid result schema type error.
        :raises InvalidResponseStatusCodeError: invalid response status code error.
        :raises InvalidRequestLimitError: invalid request limit error.
        :raises InvalidLifetimeError: invalid lifetime error.
        :raises RequestLimitOrLifetimeRequiredError: request limit or lifetime required error.
        """

        super().__init__(rule, **options)

        request_limit = options.get('request_limit')
        lifetime = options.get('lifetime')

        if request_limit is None and lifetime is None:
            raise RequestLimitOrLifetimeRequiredError('Request limit or lifetime must be '
                                                      'provided to define a temporary route.')

        if request_limit is not None and request_limit <= 0:
            raise InvalidRequestLimitError('Request limit for temporary routes '
                                           'must be a positive number.')

        if lifetime is not None and lifetime <= 0:
            raise InvalidLifetimeError('Lifetime for temporary routes must '
                                       'be a positive number of seconds.')

        self._lifetime = lifetime
        self._request_limit = request_limit
        self._registered_time = time.time()
        self._total_requests = 0

        # the threading lock must be separated for each instance.
        # because each route will have its own unique instance during application runtime.
        self._lock = Lock()

    def _is_request_limit_reached(self):
        """
        gets a value indicating that request count limit has been reached.

        :rtype: bool
        """

        if self._request_limit is not None:
            return self._total_requests >= self._request_limit

        return False

    def _is_lifetime_ended(self):
        """
        gets a value indicating that lifetime of this route has been ended.

        :rtype: bool
        """

        if self._lifetime is not None:
            return time.time() - self._registered_time >= self._lifetime

        return False

    def _should_unregister(self):
        """
        gets a value indicating that this route must be unregistered.

        :rtype: bool
        """

        return self._is_request_limit_reached() or self._is_lifetime_ended()

    def _increase_performed_requests(self):
        """
        increases performed requests count of this route by one.
        """

        with self._lock:
            self._total_requests = self._total_requests + 1

    def _handle(self, inputs, **options):
        """
        handles the current route.

        :param dict inputs: view function inputs.

        :raises URLNotFoundError: url not found error.
        """

        if self._should_unregister() is True:
            self._unregister()
            raise URLNotFoundError(_('The requested URL was not found on the server. If you '
                                     'entered the URL manually please check your spelling and '
                                     'try again.'))

        self._increase_performed_requests()

    def _finished(self, result, **options):
        """
        does the extra required works after the view function has been executed.

        this method will unregister the route if the limit or lifetime has been reached.

        :param object result: view function execution result.
        """

        if self._should_unregister() is True:
            self._unregister()

    def _unregister(self):
        """
        unregisters current route from application.
        """

        self.map.remove(self)



# -*- coding: utf-8 -*-
"""
babel handlers extract module.
"""

from pyrin.globalization.locale.babel.enumerations import BabelCLIHandlersEnum
from pyrin.globalization.locale.babel.interface import BabelCLIHandlerBase
from pyrin.globalization.locale.babel.decorators import babel_cli_handler
from pyrin.globalization.locale.babel.handlers.params import OmitHeaderParam, \
    WidthParam, NoWrapParam, CharsetParam, KeywordsParam, NoDefaultKeywordsParam, \
    MappingParam, NoLocationParam, AddLocationParam, SortOutputParam, SortByFileParam, \
    MSGIDBugsAddressParam, CopyrightHolderParam, ProjectParam, VersionParam, \
    AddCommentsParam, StripCommentsParam, InputPathsParam, OutputTemplateFileParam


@babel_cli_handler()
class ExtractCLIHandler(BabelCLIHandlerBase):
    """
    extract cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of ExtractCLIHandler.
        """

        super().__init__(BabelCLIHandlersEnum.EXTRACT)

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        :param list[CLIParamBase] params: list of all params.
        """

        params.extend([OmitHeaderParam(), WidthParam(), NoWrapParam(),
                       CharsetParam(), KeywordsParam(), NoDefaultKeywordsParam(),
                       MappingParam(), NoLocationParam(), AddLocationParam(),
                       SortOutputParam(), SortByFileParam(), MSGIDBugsAddressParam(),
                       CopyrightHolderParam(), ProjectParam(), VersionParam(),
                       AddCommentsParam(), StripCommentsParam(), InputPathsParam(),
                       OutputTemplateFileParam()])

        return super()._inject_params(params)



# -*- coding: utf-8 -*-
"""
api hooks module.
"""

from pyrin.core.structs import Hook


class APIHookBase(Hook):
    """
    api hook base class.

    all packages that need to be hooked into api business must
    implement this class and register it in api hooks.
    """

    def exception_occurred(self, error, **options):
        """
        this method will be called when an exception is occurred.

        :param Exception error: exception instance that has been occurred.
        """
        pass



# -*- coding: utf-8 -*-
"""
hashing services module.
"""

from pyrin.application.services import get_component
from pyrin.security.hashing import HashingPackage


def register_hashing_handler(instance, **options):
    """
    registers a new hashing handler or replaces the existing one
    if `replace=True` is provided. otherwise, it raises an error
    on adding an instance which it's name is already available
    in registered handlers.

    :param AbstractHashingBase instance: hashing handler to be registered.
                                         it must be an instance of AbstractHashingBase.

    :keyword bool replace: specifies that if there is another registered
                           handler with the same name, replace it with
                           the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidHashingHandlerTypeError: invalid hashing handler type error.
    :raises InvalidHashingHandlerNameError: invalid hashing handler name error.
    :raises DuplicatedHashingHandlerError: duplicated hashing handler error.
    """

    return get_component(HashingPackage.COMPONENT_NAME).register_hashing_handler(instance,
                                                                                 **options)


def generate_hash(text, **options):
    """
    gets the hash of input text using a random or specified salt.

    :param str text: text to be hashed.

    :keyword str handler_name: handler name to be used for hash generation.
                               if not provided, default handler from
                               relevant configs will be used.

    :keyword bytes salt: salt to be used for hashing.
                         if not provided, a random salt will be generated
                         considering `salt_length` option.

    :keyword str internal_algorithm: internal algorithm to be used
                                     for hashing. if not provided,
                                     default value from relevant
                                     config will be used.

    :keyword int rounds: rounds to perform for generating hash.
                         if not provided, default value from
                         relevant config will be used.

    :keyword int salt_length: salt length to be used for hashing.
                              if `salt` option is provided, then
                              this value will be ignored.
                              if not provided, default value from
                              relevant config will be used.

    :keyword str prefix: prefix to be used for bcrypt hashing.

    :rtype: str
    """

    return get_component(HashingPackage.COMPONENT_NAME).generate_hash(text, **options)


def is_match(text, full_hashed_value, **options):
    """
    gets a value indicating that given text's
    hash is identical to given full hashed value.

    :param str text: text to be hashed.
    :param str full_hashed_value: full hashed value to compare with.

    :rtype: bool
    """

    return get_component(HashingPackage.COMPONENT_NAME).is_match(text, full_hashed_value,
                                                                 **options)



# -*- coding: utf-8 -*-
"""
alembic template handler module.
"""

import os

import pyrin.application.services as application_services
import pyrin.database.services as database_services

from pyrin.core.structs import DTO
from pyrin.database.migration.alembic.enumerations import AlembicCLIHandlersEnum
from pyrin.template.decorators import template_handler
from pyrin.template.handlers.base import TemplateHandlerBase
from pyrin.database.migration.alembic import AlembicPackage


@template_handler()
class AlembicTemplateHandler(TemplateHandlerBase):
    """
    alembic template handler class.
    """

    def __init__(self):
        """
        initializes an instance of AlembicTemplateHandler.
        """

        pyrin_path = application_services.get_pyrin_main_package_path()
        target = application_services.get_migrations_path()
        source = os.path.abspath(os.path.join(pyrin_path, 'database', 'migration',
                                              'alembic', 'template', 'files'))

        super().__init__(AlembicCLIHandlersEnum.ENABLE, source, target)

    def _get_file_patterns(self):
        """
        gets the file patterns that should be included in replacement operation.

        :rtype: list[str]
        """

        return ['.py']

    def _get_data(self):
        """
        gets the data required in template generation to replace in files.

        :rtype: dict
        """

        return DTO(APPLICATION_CLASS=application_services.get_class_name(),
                   APPLICATION_MODULE=application_services.get_module_name(),
                   APPLICATION_IMPORT_NAME=application_services.get_application_name())

    def _get_config_data(self):
        """
        gets the data required in template generation to replace in config files.

        :rtype: dict
        """

        return DTO(script_location=application_services.get_migrations_path(),
                   databases=database_services.get_all_database_names())

    def _get_config_stores(self):
        """
        gets the config store names which belong to this template.

        :rtype: list[str]
        """

        return AlembicPackage.CONFIG_STORE_NAMES

    def _get_required_directories(self):
        """
        gets the required directory names that must be created in target path.

        :rtype: list[str]
        """

        return ['sql', 'versions']



# -*- coding: utf-8 -*-
"""
model decorators module.
"""

import pyrin.database.services as database_services
import pyrin.database.model.services as model_services


def bind(name, **options):
    """
    decorator to bind a model class to a database.

    :param str name: bind name to associate with the model.

    :raises InvalidEntityTypeError: invalid entity type error.

    :returns: model class.
    :rtype: type[BaseEntity]
    """

    def decorator(cls):
        """
        decorates the given model class and binds it with the specified database.

        :param type[BaseEntity] cls: model class.

        :returns: model class.
        :rtype: type[BaseEntity]
        """

        database_services.register_bind(cls, name, **options)

        return cls

    return decorator


def model_hook():
    """
    decorator to register a model hook.

    :raises InvalidModelHookTypeError: invalid model hook type error.

    :returns: model hook class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available model hooks.

        :param type cls: model hook class.

        :returns: model hook class.
        :rtype: type
        """

        instance = cls()
        model_services.register_hook(instance)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
model base module.
"""

from pyrin.core.structs import CoreObject
from pyrin.core.decorators import class_property
from pyrin.database.model.mixin import CRUDMixin, MagicMethodMixin, QueryMixin, \
    ForeignKeyMixin, ColumnMixin, PrimaryKeyMixin, RelationshipMixin, \
    HybridPropertyMixin, ConverterMixin, AttributeMixin, MetadataMixin, \
    ModelCacheMixin, DefaultPrefetchMixin, OrderingMixin


class BaseEntity(MagicMethodMixin, PrimaryKeyMixin,
                 ForeignKeyMixin, ColumnMixin,
                 RelationshipMixin, HybridPropertyMixin,
                 AttributeMixin, CRUDMixin,
                 QueryMixin, ConverterMixin,
                 MetadataMixin, ModelCacheMixin,
                 DefaultPrefetchMixin, OrderingMixin,
                 CoreObject):
    """
    base entity class.

    it should be used as the base class for every declarative base class.
    for example `CoreEntity`, which by default is the application's declarative
    base class, inherits from this.

    if you want to implement a new declarative base class for your application models
    instead of `CoreEntity`, you must inherit your new base class from `BaseEntity`.
    because application will check isinstance() on this class's type to detect models.
    the new base class must be a `DeclarativeMeta`.
    note that your application must have a unique declarative base class for all
    models, do not mix the use of your new base class and `CoreEntity`, otherwise
    you will face problems in migrations and also multi-database environments.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of BaseEntity.

        note that this method will only be called on user code, meaning
        that results returned by orm from database will not call `__init__`
        of each entity.

        it could fill primary keys, foreign keys, other columns, hybrid properties
        and also relationship properties provided in keyword arguments.
        note that relationship values must be entities. this method could
        not convert relationships which are dict, into entities.

        :keyword SECURE_TRUE | SECURE_FALSE writable: specifies that any column which has
                                                      `allow_write=False` or its name starts
                                                      with underscore `_`, should not be
                                                      populated from given values. this
                                                      is useful if you want to fill an
                                                      entity with keyword arguments passed
                                                      from client and then doing the
                                                      validation. but do not want to expose
                                                      a security risk. especially in update
                                                      operations. defaults to `SECURE_TRUE`
                                                      if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE ignore_invalid_column: specifies that if a key is
                                                                   not available in entity
                                                                   columns, do not raise an
                                                                   error. defaults to
                                                                   `SECURE_TRUE` if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE ignore_pk: specifies that any primary key column
                                                       should not be populated with given
                                                       values. this is useful if you want to
                                                       fill an entity with keyword arguments
                                                       passed from client and then doing the
                                                       validation. but do not want to let user
                                                       set primary keys and exposes a security
                                                       risk. especially in update operations.
                                                       defaults to `SECURE_TRUE` if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE ignore_fk: specifies that any foreign key column
                                                       should not be populated with given
                                                       values. this is useful if you want
                                                       to fill an entity with keyword arguments
                                                       passed from client and then doing the
                                                       validation. but do not want to let user
                                                       set foreign keys and exposes a security
                                                       risk. especially in update operations.
                                                       defaults to `SECURE_FALSE` if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE ignore_relationships: specifies that any relationship
                                                                  property should not be populated
                                                                  with given values. defaults to
                                                                  `SECURE_TRUE` if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE populate_all: specifies that all available values
                                                          must be populated from provided keyword
                                                          arguments. if set to `SECURE_TRUE`, all
                                                          other parameters will be bypassed.
                                                          this is for convenience of usage.
                                                          defaults to `SECURE_FALSE` if not
                                                          provided.

        :keyword SECURE_TRUE | SECURE_FALSE prefetch_complex_defaults: specifies that all columns
                                                                       that have default values
                                                                       for insert must fetch their
                                                                       default value if no value
                                                                       is provided for them in
                                                                       `kwargs`.
                                                                       note that scalar default
                                                                       values will always be
                                                                       fetched and this option is
                                                                       only for sequence or
                                                                       callable default values.
                                                                       defaults to `SECURE_TRUE`
                                                                       if not provided.

        :note prefetch_complex_defaults: for callable defaults we have to pass
                                         `ExecutionContext` as None because there is no
                                         such context in prefetch time. so if your callable
                                         default actually needs a valid context, you can not
                                         prefetch its value and must disable prefetching by
                                         passing `prefetch_complex_defaults=SECURE_FALSE`.
                                         otherwise unexpected behavior may occur.

        :raises ColumnNotExistedError: column not existed error.
        """

        super().__init__()

        self._set_name(self.__class__.__name__)
        self.from_dict(**kwargs)
        self.prefetch_insert_defaults(**kwargs)

    @class_property
    def _base_entity_class(cls):
        """
        gets base entity class of application.

        it should return type of `BaseEntity` class itself.
        this method is required to overcome circular dependency problem as mixin
        module could not import `BaseEntity` because `BaseEntity` itself references
        to mixin module. and also we could not inject `BaseEntity` dependency through
        `__init__()` method of `MagicMethodMixin` class, because sqlalchemy does not
        call `__init__()` method of entities for populating database results, so
        `__init__()` call is not guaranteed and will only take place on user code.
        so we have to define this method to get `BaseEntity` type here.
        and this is more beautiful than importing `BaseEntity` inside a method
        of `MagicMethodMixin` class.

        :rtype: type[BaseEntity]
        """

        return BaseEntity



# -*- coding: utf-8 -*-
"""
security exceptions module.
"""

from pyrin.core.exceptions import CoreBusinessException, CoreException


class SecurityManagerException(CoreException):
    """
    security manager exception.
    """
    pass


class SecurityManagerBusinessException(CoreBusinessException,
                                       SecurityManagerException):
    """
    security manager business exception.
    """
    pass


class InvalidPasswordLengthError(SecurityManagerBusinessException):
    """
    invalid password length error.
    """
    pass


class InvalidEncryptionTextLengthError(SecurityManagerBusinessException):
    """
    invalid encryption text length error.
    """
    pass



# -*- coding: utf-8 -*-
"""
hashing interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError


class HashingSingletonMeta(MultiSingletonMeta):
    """
    hashing singleton meta class.
    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractHashingBase(CoreObject, metaclass=HashingSingletonMeta):
    """
    abstract hashing base class.
    all application hashing handlers must be subclassed from this.
    """

    @abstractmethod
    def generate_hash(self, text, **options):
        """
        gets the hash of input text using a random or specified salt.

        :param str text: text to be hashed.

        :keyword bytes salt: salt to be used for hashing.
                             if not provided, a random salt will be generated
                             considering `salt_length` option.

        :keyword int rounds: rounds to perform for generating hash.
                             if not provided, default value from
                             relevant config will be used.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def is_match(self, text, full_hashed_value, **options):
        """
        gets a value indicating that given text's
        hash is identical to given full hashed value.

        :param str text: text to be hashed.
        :param str full_hashed_value: full hashed value to compare with.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
authentication exceptions module.
"""

from pyrin.core.enumerations import ClientErrorResponseCodeEnum
from pyrin.core.exceptions import CoreException, CoreBusinessException


class AuthenticationManagerException(CoreException):
    """
    authentication manager exception.
    """
    pass


class AuthenticationManagerBusinessException(CoreBusinessException,
                                             AuthenticationManagerException):
    """
    authentication manager business exception.
    """
    pass


class AuthenticationFailedError(AuthenticationManagerBusinessException):
    """
    authentication failed error.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of AuthenticationFailedError.

        :keyword dict data: extra data for exception.
        """

        super().__init__(*args, **kwargs)
        self._code = ClientErrorResponseCodeEnum.UNAUTHORIZED


class AccessTokenRequiredError(AuthenticationFailedError):
    """
    access token required error.
    """
    pass


class InvalidPayloadDataError(AuthenticationFailedError):
    """
    invalid payload data error.
    """
    pass



# -*- coding: utf-8 -*-
"""
orm types custom module.
"""

import uuid

from sqlalchemy.types import CHAR
from sqlalchemy import TIMESTAMP, DateTime
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.dialects.mssql import UNIQUEIDENTIFIER

from pyrin.database.enumerations import DialectEnum
from pyrin.database.orm.types.base import CoreCustomType
from pyrin.database.orm.types.mixin import DateTimeMixin


class GUID(CoreCustomType):
    """
    guid type class.

    it's a platform-independent guid type.
    uses postgresql's UUID type or mssql's UNIQUEIDENTIFIER
    type, otherwise uses CHAR(36), storing as stringified hex values.
    """

    impl = CHAR

    def _to_database(self, value, dialect):
        """
        converts given value to be emitted to database.

        :param object value: value to be processed.
        :param Dialect dialect: the dialect in use.

        :rtype: str
        """

        if dialect.name in (DialectEnum.POSTGRESQL, DialectEnum.SQLSERVER):
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return str(uuid.UUID(value))
            else:
                return str(value)

    def _from_database(self, value, dialect):
        """
        converts given value to python type after fetching it from database.

        :param object value: value to be processed.
        :param Dialect dialect: the dialect in use.

        :rtype: uuid.UUID
        """

        if not isinstance(value, uuid.UUID):
            value = uuid.UUID(value)

        return value

    def _coerce_to_string(self, value, dialect):
        """
        coerces the given value to string before sending to database.

        :param object value: value to be processed.
        :param Dialect dialect: the dialect in use.

        :rtype: str
        """

        return value

    def load_dialect_impl(self, dialect):
        """
        returns a `TypeEngine` object corresponding to a dialect.

        :param Dialect dialect: the dialect in use.

        :rtype: TypeEngine
        """

        if dialect.name == DialectEnum.POSTGRESQL:
            return dialect.type_descriptor(UUID())
        elif dialect.name == DialectEnum.SQLSERVER:
            return dialect.type_descriptor(UNIQUEIDENTIFIER())
        else:
            return dialect.type_descriptor(CHAR(36))

    def compare_against_backend(self, dialect, conn_type):
        """
        returns True if this type is the same as the given database type.

        or None to allow the default implementation to compare these
        types. a return value of False means the given type does not
        match this type.

        :param Dialect dialect: the dialect in use.
        :param TypeEngine conn_type: type of the returned value from database.

        :rtype: bool
        """

        if dialect.name == DialectEnum.POSTGRESQL:
            return isinstance(conn_type, UUID)
        elif dialect.name == DialectEnum.SQLSERVER:
            return isinstance(conn_type, UNIQUEIDENTIFIER)
        else:
            return isinstance(conn_type, CHAR)

    @property
    def python_type(self):
        """
        gets the python type object expected to be returned by instances of this type.

        :rtype: type[uuid.UUID]
        """

        return uuid.UUID


class CoreTimeStamp(DateTimeMixin):
    """
    core timestamp class.

    this is a helper type that will handle datetime values correctly on sqlite backend.
    on sqlite backend, the timezone of database will always considered as utc.
    it works as default on other backends.
    """

    impl = TIMESTAMP


class CoreDateTime(DateTimeMixin):
    """
    core datetime class.

    this is a helper type that will handle datetime values correctly on sqlite backend.
    on sqlite backend, the timezone of database will always considered as utc.
    it works as default on other backends.
    """

    impl = DateTime



# -*- coding: utf-8 -*-
"""
application mixin module.
"""

import sys
import signal

from abc import abstractmethod

import pyrin.logging.services as logging_services

from pyrin.core.structs import CoreObject
from pyrin.core.exceptions import CoreNotImplementedError


class SignalMixin(CoreObject):
    """
    signal mixin class.

    every class which needs to handle signals must inherit from this.
    note that not all operating systems support signals, so this will work on some of them.
    """

    def __init__(self, **options):
        """
        initializes an instance of SignalMixin.
        """

        super().__init__()

        # handling all signals that their default action is `terminate`,
        # to log if something goes wrong.
        try:
            signal.signal(signal.SIGABRT, self._log_signal)
            signal.signal(signal.SIGBUS, self._log_signal)
            signal.signal(signal.SIGFPE, self._log_signal)
            signal.signal(signal.SIGHUP, self._log_signal)
            signal.signal(signal.SIGILL, self._log_signal)
            signal.signal(signal.SIGPIPE, self._log_signal)
            signal.signal(signal.SIGPOLL, self._log_signal)
            signal.signal(signal.SIGPROF, self._log_signal)
            signal.signal(signal.SIGRTMIN, self._log_signal)
            signal.signal(signal.SIGRTMAX, self._log_signal)
            signal.signal(signal.SIGQUIT, self._log_signal)
            signal.signal(signal.SIGSEGV, self._log_signal)
            signal.signal(signal.SIGSYS, self._log_signal)
            signal.signal(signal.SIGTRAP, self._log_signal)
            signal.signal(signal.SIGXCPU, self._log_signal)
            signal.signal(signal.SIGXFSZ, self._log_signal)
        except Exception:
            # ignore on systems that do not support signals.
            pass

    def _log_signal(self, signal_number, frame):
        """
        logs the caught termination signal and terminates the application.

        :param int signal_number: signal number that caused termination.
        :param int | FrameType frame: interrupted stack frame.
        """

        message = 'A TERMINATION SIGNAL CAUGHT! signal [{signal_number}] ' \
                  'is going to terminate the application [{name}].' \
            .format(signal_number=signal_number,
                    name=self.get_application_name())

        logging_services.error(message)

        # terminating the application with 'status = 1' to notice a failure.
        self._terminate(signal_number, status=1)

    def _terminate(self, signal_number, **options):
        """
        terminates the application.

        this method should not be called directly.
        it is defined for cases that application has to
        be terminated for some unexpected reasons.

        :param int signal_number: signal number that caused termination.

        :keyword int status: status code to use for application exit.
                             if not provided, status=0 will be used.
        """

        message = 'Terminating application [{name}].'.format(
            name=self.get_application_name())

        logging_services.error(message)

        # forcing termination after 60 seconds if it was not done.
        signal.alarm(60)

        self._prepare_termination(signal_number)
        sys.exit(options.get('status', 0))

    @abstractmethod
    def get_application_name(self):
        """
        gets the application name.

        this method must be implemented in subclasses.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    def _prepare_termination(self, signal_number):
        """
        prepares for termination.

        this method is intended to be overridden in subclasses.

        :param int signal_number: signal number that caused termination.
        """
        pass



# -*- coding: utf-8 -*-
"""
audit component module.
"""

from pyrin.audit import AuditPackage
from pyrin.audit.manager import AuditManager
from pyrin.application.structs import Component
from pyrin.application.decorators import component


@component(AuditPackage.COMPONENT_NAME)
class AuditComponent(Component, AuditManager):
    """
    audit component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
celery component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.task_queues.celery import CeleryPackage
from pyrin.task_queues.celery.manager import CeleryManager


@component(CeleryPackage.COMPONENT_NAME)
class CeleryComponent(Component, CeleryManager):
    """
    celery component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
template handlers package module.
"""

import os

import pyrin.utils.path as path_utils
import pyrin.application.services as application_services

from pyrin.core.structs import DTO
from pyrin.template.decorators import template_handler
from pyrin.template.enumerations import TemplateCLIHandlersEnum
from pyrin.template.handlers.base import TemplateHandlerWithInterfaceInputBase
from pyrin.template.handlers.exceptions import InvalidPackagePathError, \
    InvalidPackageClassNameError


class PackageTemplateHandlerBase(TemplateHandlerWithInterfaceInputBase):
    """
    package template handler base class.
    """

    def __init__(self, name, source):
        """
        initializes an instance of PackageTemplateHandlerBase.

        :param str name: name of the handler.
                         each handler must have a unique name.

        :param str source: source directory of template files.
        """

        self._package_path = None
        self._package_full_path = None
        self._package_name = None
        self._package_full_name = None
        self._package_title = None
        self._package_alias = None
        self._package_class_name = None
        self._application_path = application_services.get_application_main_package_path()
        self._working_directory = application_services.get_working_directory()

        super().__init__(name, source)

    def _show_interface(self, package_path=None, package_class_name=None):
        """
        shows cli prompt to get inputs from user.

        :param str package_path: the new package path. it must be a relative
                                 path inside application main package path.

        :param str package_class_name: the new package class name.

        :raises InvalidPackagePathError: invalid package path error.
        :raises InvalidPackageClassNameError: invalid package class name error.
        """

        if package_path is None:
            package_path = input('Please input the new package relative path: ')

        if package_class_name is None:
            package_class_name = input('Please input the new package class name: ')

        self._set_attributes(package_path, package_class_name)

    def _validate_inputs(self, package_path, package_class_name):
        """
        validates the inputs to be used by this handler.

        :param str package_path: the new package path. it must be a relative
                                 path inside application main package path.

        :param str package_class_name: the new package class name.

        :raises InvalidPackagePathError: invalid package path error.
        :raises InvalidPackageClassNameError: invalid package class name error.
        """

        if package_path in (None, '') or package_path.isspace() or \
                os.path.isabs(package_path):
            raise InvalidPackagePathError('New package path is invalid.')

        if package_class_name in (None, '') or package_class_name.isspace():
            raise InvalidPackageClassNameError('New package class name is invalid.')

    def _set_attributes(self, package_path, package_class_name):
        """
        sets the required attributes based on given inputs.

        :param str package_path: the new package path. it must be a relative
                                 path inside application main package path.

        :param str package_class_name: the new package class name.

        :raises InvalidPackagePathError: invalid package path error.
        :raises InvalidPackageClassNameError: invalid package class name error.
        """

        self._validate_inputs(package_path, package_class_name)

        self._package_path = package_path.rstrip(os.path.sep).rstrip(
            os.path.altsep).replace(' ', '').lower()
        self._package_full_path = os.path.abspath(os.path.join(
            self._application_path, self._package_path))
        self._package_title = ' '.join(self._package_path.split(os.path.sep)).lower().strip()
        self._package_alias = '_'.join(self._package_path.split(os.path.sep)).lower().strip()
        self._package_name = '.'.join(self._package_path.split(os.path.sep)).lower().strip()
        self._package_full_name = path_utils.get_package_name(self._package_full_path,
                                                              self._working_directory)
        self._package_class_name = package_class_name.replace(' ', '')
        self._target = self._package_full_path

    def _get_file_patterns(self):
        """
        gets the file patterns that should be included in replacement operation.

        :rtype: list[str]
        """

        return ['.py']

    def _get_data(self):
        """
        gets the data required in template generation to replace in files.

        :rtype: dict
        """

        return DTO(PACKAGE_NAME=self._package_name,
                   PACKAGE_TITLE=self._package_title,
                   PACKAGE_ALIAS=self._package_alias,
                   PACKAGE_CLASS_NAME=self._package_class_name,
                   PACKAGE_FULL_NAME=self._package_full_name)


@template_handler()
class PackageTemplateHandler(PackageTemplateHandlerBase):
    """
    package template handler class.

    this template handler will be used to create new application packages.
    """

    def __init__(self):
        """
        initializes an instance of PackageTemplateHandler.
        """

        pyrin_path = application_services.get_pyrin_main_package_path()
        source = os.path.abspath(os.path.join(pyrin_path, 'template', 'files', 'package'))

        super().__init__(TemplateCLIHandlersEnum.PACKAGE, source)


@template_handler()
class EmptyPackageTemplateHandler(PackageTemplateHandlerBase):
    """
    empty package template handler class.

    this template handler will be used to create new empty application packages.
    """

    def __init__(self):
        """
        initializes an instance of EmptyPackageTemplateHandler.
        """

        pyrin_path = application_services.get_pyrin_main_package_path()
        source = os.path.abspath(os.path.join(pyrin_path, 'template', 'files', 'empty_package'))

        super().__init__(TemplateCLIHandlersEnum.EMPTY_PACKAGE, source)



# -*- coding: utf-8 -*-
"""
session exceptions module.
"""

from pyrin.core.exceptions import CoreBusinessException, CoreException


class SessionManagerException(CoreException):
    """
    session manager exception.
    """
    pass


class SessionManagerBusinessException(CoreBusinessException,
                                      SessionManagerException):
    """
    session manager business exception.
    """
    pass


class CouldNotOverwriteCurrentUserError(SessionManagerException):
    """
    could not overwrite current user error.
    """
    pass


class InvalidUserError(SessionManagerException):
    """
    invalid user error.
    """
    pass


class InvalidComponentCustomKeyError(SessionManagerException):
    """
    invalid component custom key error.
    """
    pass



# -*- coding: utf-8 -*-
"""
serializer services module.
"""

from pyrin.converters.serializer import SerializerPackage
from pyrin.application.services import get_component


def serialize(value, **options):
    """
    serializes the given value.

    returns serialized value on success or the
    same input value if serialization fails.

    all extra keyword arguments will be passed to underlying serializer.

    :param object value: value to be serialized.

    :keyword ResultSchema result_schema: result schema instance to be
                                         used to create computed columns.
                                         defaults to None if not provided.

    :returns: serialized object
    :rtype: dict | list[dict]
    """

    return get_component(SerializerPackage.COMPONENT_NAME).serialize(value, **options)


def register_serializer(instance, **options):
    """
    registers a new serializer or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding a serializer which is already registered.

    :param AbstractSerializerBase instance: serializer to be registered.
                                            it must be an instance of
                                            AbstractSerializerBase.

    :keyword bool replace: specifies that if there is another registered
                           serializer with the same accepted type,
                           replace it with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidSerializerTypeError: invalid serializer type error.
    :raises DuplicatedSerializerError: duplicated serializer error.
    """

    return get_component(SerializerPackage.COMPONENT_NAME).register_serializer(instance,
                                                                               **options)


def get_serializers(accepted_type):
    """
    gets the registered serializers for given type.

    it returns an empty list if no serializer found for given type.

    :param type accepted_type: gets the serializers which are
                               registered for the accepted type.

    :rtype: list[AbstractSerializerBase]
    """

    return get_component(SerializerPackage.COMPONENT_NAME).get_serializers(accepted_type)



# -*- coding: utf-8 -*-
"""
deserializer manager module.
"""

from pyrin.converters.deserializer import DeserializerPackage
from pyrin.converters.deserializer.interface import AbstractDeserializerBase
from pyrin.core.structs import Context, Manager
from pyrin.core.globals import NULL
from pyrin.utils.custom_print import print_warning
from pyrin.converters.deserializer.exceptions import InvalidDeserializerTypeError, \
    DuplicatedDeserializerError


class DeserializerManager(Manager):
    """
    deserializer manager class.
    """

    package_class = DeserializerPackage

    def __init__(self):
        """
        initializes an instance of DeserializerManager.
        """

        super().__init__()

        # a dictionary containing information of registered deserializers.
        # example: dict(type accepted_type: list[AbstractDeserializerBase] instances)
        self._deserializers = Context()

    def deserialize(self, value, **options):
        """
        deserializes the given value.

        returns deserialized object on success or returns
        the same input value if deserialization fails.

        :param object value: value to be deserialized.

        :keyword bool include_internal: specifies that any chained internal deserializer
                                        must also be used for deserialization. if set to
                                        False, only non-internal deserializers will be used.
                                        defaults to True if not provided.

        :returns: deserialized object
        """

        options.update(accepted_type=type(value))
        for deserializer in self.get_deserializers(**options):
            deserialized_value = deserializer.deserialize(value, **options)

            if deserialized_value is not NULL:
                return deserialized_value
            continue

        return value

    def register_deserializer(self, instance, **options):
        """
        registers a new deserializer or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding a deserializer which is already registered.

        :param AbstractDeserializerBase instance: deserializer to be registered.
                                                  it must be an instance of
                                                  AbstractDeserializerBase.

        :keyword bool replace: specifies that if there is another registered
                               deserializer with the same name and accepted type,
                               replace it with the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidDeserializerTypeError: invalid deserializer type error.
        :raises DuplicatedDeserializerError: duplicated deserializer error.
        """

        if not isinstance(instance, AbstractDeserializerBase):
            raise InvalidDeserializerTypeError('Input parameter [{instance}] is '
                                               'not an instance of [{base}].'
                                               .format(instance=instance,
                                                       base=AbstractDeserializerBase))

        previous_instances = self._deserializers.get(instance.accepted_type, [])
        if instance.accepted_type in self._deserializers:
            if len(previous_instances) > 0:
                old_instance = self._get_deserializer_with_name(instance.get_name(),
                                                                previous_instances)
                if old_instance is not None:
                    replace = options.get('replace', False)
                    if replace is not True:
                        raise DuplicatedDeserializerError('There is another registered '
                                                          'deserializer with name [{name}] '
                                                          'for accepted type [{accepted_type}] '
                                                          'but "replace" option is not set, so '
                                                          'deserializer [{instance}] could not '
                                                          'be registered.'
                                                          .format(name=instance.get_name(),
                                                                  accepted_type=instance.
                                                                  accepted_type,
                                                                  instance=instance))

                    print_warning('Deserializer [{old_instance}] is going '
                                  'to be replaced by [{new_instance}].'
                                  .format(old_instance=old_instance,
                                          new_instance=instance))

                    previous_instances.remove(old_instance)

        previous_instances.append(instance)
        self._set_next_handlers(previous_instances)
        self._deserializers[instance.accepted_type] = previous_instances

    def _set_next_handlers(self, deserializers):
        """
        sets next handler for each deserializer in the input list.

        :param list[AbstractDeserializerBase] deserializers: list of deserializers.
        """

        length = len(deserializers)
        for i in range(length):
            if i == length - 1:
                deserializers[i].set_next(None)
            else:
                deserializers[i].set_next(deserializers[i + 1])

    def _get_deserializer_with_name(self, name, deserializers):
        """
        gets a deserializer with the given name from input deserializers list.

        if not available, it returns None.

        :param str name: deserializer name to get its instance.
        :param list[AbstractDeserializerBase] deserializers: list of deserializers.

        :raises DuplicatedDeserializerError: duplicated deserializer error.

        :rtype: AbstractDeserializerBase
        """

        result = [item for item in deserializers if item.get_name() == name]

        if result is None or len(result) <= 0:
            return None

        if len(result) > 1:
            raise DuplicatedDeserializerError('There are multiple deserializers with '
                                              'name [{name}]. it could be due to a '
                                              'bug in registering deserializers.'
                                              .format(name=name))
        if len(result) == 1:
            return result[0]

    def get_deserializers(self, **options):
        """
        gets all registered deserializers.

        it could filter deserializers for a specific type if provided.
        it only returns the first deserializer for each type, because
        all deserializers for a given type, are chained together.

        :keyword type accepted_type: specifies to get deserializers which are registered for
                                     the accepted type. if not provided, all deserializers
                                     will be returned.

        :rtype: list[AbstractDeserializerBase]
        """

        accepted_type = options.get('accepted_type', None)

        if accepted_type is None:
            all_deserializers = []
            for deserializer_type in self._deserializers:
                all_deserializers.append(self._deserializers[deserializer_type][0])
            return all_deserializers
        else:
            deserializer_keys = [key for key in self._deserializers
                                 if issubclass(accepted_type, key)]

            specific_deserializers = []
            for key in deserializer_keys:
                specific_deserializers.append(self._deserializers[key][0])

            return specific_deserializers



# -*- coding: utf-8 -*-
"""
router exceptions module.
"""

from pyrin.core.exceptions import CoreException


class RouterException(CoreException):
    """
    router exception.
    """
    pass


class RouteAuthenticationMismatchError(RouterException):
    """
    route authentication mismatch error.
    """
    pass


class InvalidCustomRouteTypeError(RouterException):
    """
    invalid custom route type error.
    """
    pass



# -*- coding: utf-8 -*-
"""
database bulk component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.database.bulk import DatabaseBulkPackage
from pyrin.database.bulk.manager import DatabaseBulkManager


@component(DatabaseBulkPackage.COMPONENT_NAME)
class DatabaseBulkComponent(Component, DatabaseBulkManager):
    """
    database bulk component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
sentry services module.
"""

from pyrin.application.services import get_component
from pyrin.logging.sentry import SentryPackage


def register_integration(instance, **options):
    """
    registers a new integration or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding an integration which is already registered.

    :param AbstractSentryIntegration instance: integration instance to be registered.

    :keyword bool replace: specifies that if there is another registered
                           integration with the same name, replace it with the
                           new one, otherwise raise an error. defaults to False.

    :raises InvalidSentryIntegrationTypeError: invalid sentry integration type error.
    :raises DuplicatedSentryIntegrationError: duplicated sentry integration error.
    """

    get_component(SentryPackage.COMPONENT_NAME).register_integration(instance, **options)


def configure(**options):
    """
    configures sentry client and all enabled integrations.
    """

    get_component(SentryPackage.COMPONENT_NAME).configure(**options)


def try_report_exception(error, **options):
    """
    tries to report exception to sentry server.

    if logging integration is not enabled to send it automatically.

    :param Exception error: exception instance that has been occurred.
    """

    get_component(SentryPackage.COMPONENT_NAME).try_report_exception(error, **options)



# -*- coding: utf-8 -*-
"""
celery cli manager module.
"""

from pyrin.cli.mixin.handler import CLIMixin
from pyrin.core.structs import Manager
from pyrin.task_queues.celery.cli import CeleryCLIPackage
from pyrin.task_queues.celery.cli.interface import CeleryCLIHandlerBase


class CeleryCLIManager(Manager, CLIMixin):
    """
    celery cli manager class.
    """

    _cli_handler_type = CeleryCLIHandlerBase
    package_class = CeleryCLIPackage



# -*- coding: utf-8 -*-
"""
packaging decorators module.
"""

import pyrin.packaging.services as packaging_services


def packaging_hook():
    """
    decorator to register a packaging hook.

    :raises InvalidPackagingHookTypeError: invalid packaging hook type error.

    :returns: packaging hook class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available packaging hooks.

        :param type cls: packaging hook class.

        :returns: packaging hook class.
        :rtype: type
        """

        instance = cls()
        packaging_services.register_hook(instance)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
cors exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class CORSException(CoreException):
    """
    cors exception.
    """
    pass


class CORSBusinessException(CoreBusinessException,
                            CORSException):
    """
    cors business exception.
    """
    pass


class CORSAllowedHeadersModificationError(CORSException):
    """
    cors allowed headers modification error.
    """
    pass



# -*- coding: utf-8 -*-
"""
response status manager module.
"""

import pyrin.security.session.services as session_services

from pyrin.processor.response.status import ResponseStatusPackage
from pyrin.settings.static import DEFAULT_STATUS_CODE
from pyrin.core.structs import Manager, DTO
from pyrin.core.enumerations import SuccessfulResponseCodeEnum, HTTPMethodEnum, \
    ClientErrorResponseCodeEnum, ServerErrorResponseCodeEnum, InformationResponseCodeEnum, \
    RedirectionResponseCodeEnum


class ResponseStatusManager(Manager):
    """
    response status manager class.
    """

    # these values will be used to detect response status type from
    # status code if `strict_status=False` is passed to corresponding method.
    # if you want to provide custom non-standard response codes in your
    # application, you should override these values in a subclass.
    CLIENT_ERROR_CODE_MIN = ClientErrorResponseCodeEnum.BAD_REQUEST
    CLIENT_ERROR_CODE_MAX = 499

    SERVER_ERROR_CODE_MIN = ServerErrorResponseCodeEnum.INTERNAL_SERVER_ERROR
    SERVER_ERROR_CODE_MAX = 599

    INFORMATION_CODE_MIN = InformationResponseCodeEnum.CONTINUE
    INFORMATION_CODE_MAX = 199

    SUCCESS_CODE_MIN = SuccessfulResponseCodeEnum.OK
    SUCCESS_CODE_MAX = 299

    REDIRECTION_CODE_MIN = RedirectionResponseCodeEnum.MULTIPLE_CHOICE
    REDIRECTION_CODE_MAX = 399

    package_class = ResponseStatusPackage

    def __init__(self):
        """
        initializes an instance of ResponseStatusManager.
        """

        super().__init__()

        # a dict containing the mapping between http methods and status codes.
        # in the form of: {str http_method: int status_code}
        self.__method_to_status_map = DTO()
        self.__initialize_mapping()

    def __initialize_mapping(self):
        """
        initializes the mapping between http methods and status codes.
        """

        mapping = DTO()
        mapping[HTTPMethodEnum.GET] = SuccessfulResponseCodeEnum.OK
        mapping[HTTPMethodEnum.PUT] = SuccessfulResponseCodeEnum.OK
        mapping[HTTPMethodEnum.PATCH] = SuccessfulResponseCodeEnum.OK
        mapping[HTTPMethodEnum.DELETE] = SuccessfulResponseCodeEnum.OK
        mapping[HTTPMethodEnum.POST] = SuccessfulResponseCodeEnum.CREATED

        self._customize_mapping(mapping)
        self.__method_to_status_map = mapping

    def _customize_mapping(self, mapping):
        """
        customizes the mapping between http methods and status codes.

        this method is intended to be overridden by subclasses.
        it should always do the required modifications into the input dict in-place.

        :param dict[str, int] mapping: the mapping dict.
        :note mapping: dict[str http_method: int status_code]
        """
        pass

    def get_status_code(self, method=None, **options):
        """
        gets the corresponding status code for given http method.

        if method is not provided, it gets the status code for current request's method.

        :param str method: http method to get its corresponding status code.
        :enum method:
            GET = 'GET'
            POST = 'POST'
            PUT = 'PUT'
            DELETE = 'DELETE'
            HEAD = 'HEAD'
            TRACE = 'TRACE'
            OPTIONS = 'OPTIONS'
            PATCH = 'PATCH'
            COPY = 'COPY'
            LINK = 'LINK'
            UNLINK = 'UNLINK'
            PURGE = 'PURGE'
            VIEW = 'VIEW'

        :keyword bool has_content: specifies that the response body has content.
                                   if set to False, the status code returned will
                                   be `NO_CONTENT=204` for all http methods.
                                   defaults to True if not provided.

        :rtype: int
        """

        has_content = options.get('has_content', True)
        if has_content is False:
            return SuccessfulResponseCodeEnum.NO_CONTENT

        if method is None:
            method = session_services.get_current_request().method

        status = self.__method_to_status_map.get(method, None)
        return status or DEFAULT_STATUS_CODE

    def is_client_error(self, status_code, **options):
        """
        gets a value indicating that given status code is a client error.

        if returns True if the provided status code is
        from `ClientErrorResponseCodeEnum` values.

        :param int status_code: status code to be checked.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as client error if it is from
                                     `ClientErrorResponseCodeEnum` values. otherwise
                                     all codes from `CLIENT_ERROR_CODE_MIN` to
                                     `CLIENT_ERROR_CODE_MAX` will be considered as
                                     client error. defaults to True if not provided.

        :rtype: bool
        """

        strict_status = options.get('strict_status', True)
        if strict_status is False:
            return self.CLIENT_ERROR_CODE_MIN <= status_code <= self.CLIENT_ERROR_CODE_MAX

        return status_code in ClientErrorResponseCodeEnum.values()

    def is_server_error(self, status_code, **options):
        """
        gets a value indicating that given status code is a server error.

        if returns True if the provided status code is
        from `ServerErrorResponseCodeEnum` values.

        :param int status_code: status code to be checked.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as server error if it is from
                                     `ServerErrorResponseCodeEnum` values. otherwise
                                     all codes from `SERVER_ERROR_CODE_MIN` to
                                     `SERVER_ERROR_CODE_MAX` will be considered
                                     as server error. defaults to True if not provided.

        :rtype: bool
        """

        strict_status = options.get('strict_status', True)
        if strict_status is False:
            return self.SERVER_ERROR_CODE_MIN <= status_code <= self.SERVER_ERROR_CODE_MAX

        return status_code in ServerErrorResponseCodeEnum.values()

    def is_error(self, status_code, **options):
        """
        gets a value indicating that given status code is a client or server error.

        if returns True if the provided status code is from
        `ClientErrorResponseCodeEnum` or `ServerErrorResponseCodeEnum` values.

        :param int status_code: status code to be checked.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as error if it is from
                                     `ClientErrorResponseCodeEnum` or
                                     `ServerErrorResponseCodeEnum` values. otherwise
                                     all codes from `CLIENT_ERROR_CODE_MIN` to
                                     `CLIENT_ERROR_CODE_MAX` or from
                                     `SERVER_ERROR_CODE_MIN` to `SERVER_ERROR_CODE_MAX`
                                     will be considered as error.
                                     defaults to True if not provided.

        :rtype: bool
        """

        return self.is_server_error(status_code, **options) or \
            self.is_client_error(status_code, **options)

    def is_information(self, status_code, **options):
        """
        gets a value indicating that given status code is a information code.

        if returns True if the provided status code is
        from `InformationResponseCodeEnum` values.

        :param int status_code: status code to be checked.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as information if it is from
                                     `InformationResponseCodeEnum` values. otherwise
                                     all codes from `INFORMATION_CODE_MIN` to
                                     `INFORMATION_CODE_MAX` will be considered as
                                     information. defaults to True if not provided.

        :rtype: bool
        """

        strict_status = options.get('strict_status', True)
        if strict_status is False:
            return self.INFORMATION_CODE_MIN <= status_code <= self.INFORMATION_CODE_MAX

        return status_code in InformationResponseCodeEnum.values()

    def is_success(self, status_code, **options):
        """
        gets a value indicating that given status code is a success code.

        if returns True if the provided status code is
        from `SuccessfulResponseCodeEnum` values.

        :param int status_code: status code to be checked.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as success if it is from
                                     `SuccessfulResponseCodeEnum` values. otherwise
                                     all codes from `SUCCESS_CODE_MIN` to
                                     `SUCCESS_CODE_MAX` will be considered as success.
                                     defaults to True if not provided.

        :rtype: bool
        """

        strict_status = options.get('strict_status', True)
        if strict_status is False:
            return self.SUCCESS_CODE_MIN <= status_code <= self.SUCCESS_CODE_MAX

        return status_code in SuccessfulResponseCodeEnum.values()

    def is_redirection(self, status_code, **options):
        """
        gets a value indicating that given status code is a redirection code.

        if returns True if the provided status code is
        from `RedirectionResponseCodeEnum` values.

        :param int status_code: status code to be checked.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as redirection if it is from
                                     `RedirectionResponseCodeEnum` values. otherwise
                                     all codes from `REDIRECTION_CODE_MIN` to
                                     `REDIRECTION_CODE_MAX` will be considered
                                     as redirection. defaults to True if not provided.

        :rtype: bool
        """

        strict_status = options.get('strict_status', True)
        if strict_status is False:
            return self.REDIRECTION_CODE_MIN <= status_code <= self.REDIRECTION_CODE_MAX

        return status_code in RedirectionResponseCodeEnum.values()

    def is_processed(self, status_code, **options):
        """
        gets a value indicating that given status code is a processed code.

        processed codes are information or success or redirection code.
        if returns True if the provided status code is
        from `InformationResponseCodeEnum` or `SuccessfulResponseCodeEnum`
        or `RedirectionResponseCodeEnum` values.

        :param int status_code: status code to be checked.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as processed if it is from
                                     `InformationResponseCodeEnum` or
                                     `SuccessfulResponseCodeEnum` or
                                     `RedirectionResponseCodeEnum` values. otherwise
                                     all codes from `INFORMATION_CODE_MIN`
                                     to `INFORMATION_CODE_MAX` or from
                                     `SUCCESS_CODE_MIN` to `SUCCESS_CODE_MAX`
                                     or from `REDIRECTION_CODE_MIN` to
                                     `REDIRECTION_CODE_MAX` will be considered
                                     as processed. defaults to True if not provided.

        :rtype: bool
        """

        return self.is_success(status_code, **options) or \
            self.is_information(status_code, **options) or \
            self.is_redirection(status_code, **options)



# -*- coding: utf-8 -*-
"""
validator auto component module.
"""

from pyrin.application.decorators import component
from pyrin.validator.auto import ValidatorAutoPackage
from pyrin.validator.auto.manager import ValidatorAutoManager
from pyrin.application.structs import Component


@component(ValidatorAutoPackage.COMPONENT_NAME)
class ValidatorAutoComponent(Component, ValidatorAutoManager):
    """
    validator auto component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
response wrappers exceptions module.
"""

from pyrin.core.exceptions import CoreException


class ResponseWrappersException(CoreException):
    """
    response wrappers exception.
    """
    pass


class InvalidResponseContextKeyNameError(ResponseWrappersException):
    """
    invalid response context key name error.
    """
    pass


class ResponseContextKeyIsAlreadyPresentError(ResponseWrappersException):
    """
    response context key is already present error.
    """
    pass


class ResponseEnvironRequiredError(ResponseWrappersException):
    """
    response environ required error.
    """
    pass



# -*- coding: utf-8 -*-
"""
database bulk manager module.
"""

import math

import pyrin.converters.serializer.services as serializer_services

from pyrin.database.bulk import DatabaseBulkPackage
from pyrin.core.structs import Manager
from pyrin.database.services import get_current_store


class DatabaseBulkManager(Manager):
    """
    database bulk manager class.
    """

    package_class = DatabaseBulkPackage

    def insert(self, *entities, **options):
        """
        bulk inserts the given entities.

        note that entities must be from the same type.

        :param BaseEntity entities: entities to be inserted.

        :keyword int chunk_size: chunk size to insert values.
                                 after each chunk, store will be committed.
                                 if not provided, all values will be inserted
                                 in a single call and no commit will occur.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults to
                                                      `SECURE_TRUE` if not provided.

        :keyword dict[str, list[str]] | list[str] columns: column names to be included in result.
                                                           it could be a list of column names.
                                                           for example:
                                                           `columns=['id', 'name', 'age']`
                                                           but if you want to include
                                                           relationships, then columns for each
                                                           entity must be provided in a key for
                                                           that entity class name.
                                                           for example if there is `CarEntity` and
                                                           `PersonEntity`, it should be like this:
                                                           `columns=dict(CarEntity=
                                                                         ['id', 'name'],
                                                                         PersonEntity=
                                                                         ['id', 'age'])`
                                                           if provided column names are not
                                                           available in result, they will
                                                           be ignored.

        :note columns: dict[str entity_class_name, list[str column_name]] | list[str column_name]

        :keyword dict[str, dict[str, str]] | dict[str, str] rename: column names that must be
                                                                    renamed in the result.
                                                                    it could be a dict with keys
                                                                    as original column names and
                                                                    values as new column names
                                                                    that should be exposed instead
                                                                    of original column names.
                                                                    for example:
                                                                    `rename=dict(age='new_age',
                                                                                 name='new_name')`
                                                                    but if you want to include
                                                                    relationships, then you must
                                                                    provide a dict containing
                                                                    entity class name as key and
                                                                    for value, another dict
                                                                    containing original column
                                                                    names as keys, and column
                                                                    names that must be exposed
                                                                    instead of original names,
                                                                    as values. for example
                                                                    if there is `CarEntity` and `
                                                                    PersonEntity`, it should be
                                                                    like this:
                                                                    `rename=
                                                                    dict(CarEntity=
                                                                         dict(name='new_name'),
                                                                         PersonEntity=
                                                                         dict(age='new_age')`
                                                                    then, the value of `name`
                                                                    column in result will be
                                                                    returned as `new_name` column.
                                                                    and also value of `age` column
                                                                    in result will be returned as
                                                                    'new_age' column. if provided
                                                                    rename columns are not
                                                                    available in result, they
                                                                    will be ignored.

        :note rename: dict[str entity_class_name, dict[str original_column, str new_column]] |
                      dict[str original_column, str new_column]

        :keyword dict[str, list[str]] | list[str] exclude: column names to be excluded from
                                                           result. it could be a list of column
                                                           names. for example:
                                                           `exclude=['id', 'name', 'age']`
                                                           but if you want to include
                                                           relationships, then columns for each
                                                           entity must be provided in a key for
                                                           that entity class name.
                                                           for example if there is `CarEntity`
                                                           and `PersonEntity`, it should be
                                                           like this:
                                                           `exclude=dict(CarEntity=
                                                                         ['id', 'name'],
                                                                         PersonEntity=
                                                                         ['id', 'age'])`
                                                            if provided excluded columns are not
                                                            available in result, they will be
                                                            ignored.

        :note exclude: dict[str entity_class_name, list[str column_name]] | list[str column_name]

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.
        """

        size = len(entities)
        if size > 0:
            store = get_current_store()
            chunk = options.get('chunk_size', None)
            entity_type = type(entities[0])
            serialized_values = serializer_services.serialize(entities, **options)
            if chunk is None:
                chunk = 0
            chunk = int(chunk)
            if size <= chunk or chunk <= 0:
                store.bulk_insert_mappings(entity_type, serialized_values)
                if chunk > 0:
                    store.commit()
            else:
                rounds = math.ceil(size / chunk)
                for i in range(rounds):
                    ready = serialized_values[i * chunk:chunk * (i + 1)]
                    store.bulk_insert_mappings(entity_type, ready)
                    store.commit()

    def update(self, *entities, **options):
        """
        bulk updates the given entities.

        note that entities must be from the same type.

        :param BaseEntity entities: entities to be updated.

        :keyword int chunk_size: chunk size to update values.
                                 after each chunk, store will be committed.
                                 if not provided, all values will be updated
                                 in a single call and no commit will occur.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults to
                                                      `SECURE_TRUE` if not provided.

        :keyword dict[str, list[str]] | list[str] columns: column names to be included in result.
                                                           it could be a list of column names.
                                                           for example:
                                                           `columns=['id', 'name', 'age']`
                                                           but if you want to include
                                                           relationships, then columns for each
                                                           entity must be provided in a key for
                                                           that entity class name.
                                                           for example if there is `CarEntity` and
                                                           `PersonEntity`, it should be like this:
                                                           `columns=dict(CarEntity=
                                                                         ['id', 'name'],
                                                                         PersonEntity=
                                                                         ['id', 'age'])`
                                                           if provided column names are not
                                                           available in result, they will
                                                           be ignored.

        :note columns: dict[str entity_class_name, list[str column_name]] | list[str column_name]

        :keyword dict[str, dict[str, str]] | dict[str, str] rename: column names that must be
                                                                    renamed in the result.
                                                                    it could be a dict with keys
                                                                    as original column names and
                                                                    values as new column names
                                                                    that should be exposed instead
                                                                    of original column names.
                                                                    for example:
                                                                    `rename=dict(age='new_age',
                                                                                 name='new_name')`
                                                                    but if you want to include
                                                                    relationships, then you must
                                                                    provide a dict containing
                                                                    entity class name as key and
                                                                    for value, another dict
                                                                    containing original column
                                                                    names as keys, and column
                                                                    names that must be exposed
                                                                    instead of original names,
                                                                    as values. for example
                                                                    if there is `CarEntity` and `
                                                                    PersonEntity`, it should be
                                                                    like this:
                                                                    `rename=
                                                                    dict(CarEntity=
                                                                         dict(name='new_name'),
                                                                         PersonEntity=
                                                                         dict(age='new_age')`
                                                                    then, the value of `name`
                                                                    column in result will be
                                                                    returned as `new_name` column.
                                                                    and also value of `age` column
                                                                    in result will be returned as
                                                                    'new_age' column. if provided
                                                                    rename columns are not
                                                                    available in result, they
                                                                    will be ignored.

        :note rename: dict[str entity_class_name, dict[str original_column, str new_column]] |
                      dict[str original_column, str new_column]

        :keyword dict[str, list[str]] | list[str] exclude: column names to be excluded from
                                                           result. it could be a list of column
                                                           names. for example:
                                                           `exclude=['id', 'name', 'age']`
                                                           but if you want to include
                                                           relationships, then columns for each
                                                           entity must be provided in a key for
                                                           that entity class name.
                                                           for example if there is `CarEntity`
                                                           and `PersonEntity`, it should be
                                                           like this:
                                                           `exclude=dict(CarEntity=
                                                                         ['id', 'name'],
                                                                         PersonEntity=
                                                                         ['id', 'age'])`
                                                            if provided excluded columns are not
                                                            available in result, they will be
                                                            ignored.

        :note exclude: dict[str entity_class_name, list[str column_name]] | list[str column_name]

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.
        """

        size = len(entities)
        if size > 0:
            store = get_current_store()
            chunk = options.get('chunk_size', None)
            entity_type = type(entities[0])
            serialized_values = serializer_services.serialize(entities, **options)
            if chunk is None:
                chunk = 0
            chunk = int(chunk)
            if size <= chunk or chunk <= 0:
                store.bulk_update_mappings(entity_type, serialized_values)
                if chunk > 0:
                    store.commit()
            else:
                rounds = math.ceil(size / chunk)
                for i in range(rounds):
                    ready = serialized_values[i * chunk:chunk * (i + 1)]
                    store.bulk_update_mappings(entity_type, ready)
                    store.commit()



# -*- coding: utf-8 -*-
"""
encryption exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class EncryptionManagerException(CoreException):
    """
    encryption manager exception.
    """
    pass


class EncryptionManagerBusinessException(CoreBusinessException,
                                         EncryptionManagerException):
    """
    encryption manager business exception.
    """
    pass


class DecryptionError(EncryptionManagerBusinessException):
    """
    decryption error.
    """
    pass


class InvalidEncryptionHandlerTypeError(EncryptionManagerException):
    """
    invalid encryption handler type error.
    """
    pass


class DuplicatedEncryptionHandlerError(EncryptionManagerException):
    """
    duplicated encryption handler error.
    """
    pass


class EncryptionHandlerNotFoundError(EncryptionManagerException):
    """
    encryption handler not found error.
    """
    pass


class InvalidEncryptionHandlerNameError(EncryptionManagerException):
    """
    invalid encryption handler name error.
    """
    pass


class InvalidEncryptedValueError(EncryptionManagerBusinessException):
    """
    invalid encrypted value error.
    """
    pass



# -*- coding: utf-8 -*-
"""
utils file module.
"""

import re
import os

import pyrin.utils.path as path_utils

from pyrin.utils.exceptions import IsNotDirectoryError


def replace_file_values_regex(source, data):
    """
    replaces the values in given file with values available in given dict.

    the replacement is done using regular expression.

    :param str source: file path to replace its values.
                       it must be an absolute path.

    :param dict[str, str] data: a dict containing all values that must be
                                replaced in given file. keys of the dict
                                must be strings representing regular
                                expressions and their values must be
                                the values that should replace them.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    :raises PathNotExistedError: path not existed error.
    """

    path_utils.assert_exists(source)
    if data is not None and len(data) > 0:
        with open(source, 'r') as file:
            file_data = file.read()

        for regex, value in data.items():
            file_data = re.sub(regex, value, file_data, flags=re.MULTILINE)

        with open(source, 'w') as file:
            file.write(file_data)


def replace_files_values_regex(source, data, *patterns):
    """
    replaces the values in all files with values available in given dict.

    if patterns is given, only matching files will be included.

    :param str source: directory path to process its files.
                       the operation will also include all
                       files of all subdirectories.
                       it must be an absolute path.

    :param dict[str, str] data: a dict containing all values that must be
                                replaced in given file. keys of the dict
                                must be strings representing regular
                                expressions and their values must be
                                the values that should replace them.

    :param str patterns: file name end patterns to be included
                         in replacement operation. for example
                         it could be `'.py', '.html'`.
                         all files will be included if not provided.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    :raises PathNotExistedError: path not existed error.
    :raises IsNotDirectoryError: is not directory error.
    """

    path_utils.assert_exists(source)
    if not os.path.isdir(source):
        raise IsNotDirectoryError('Provided path [{source}] is not a directory.'
                                  .format(source=source))

    for path, directories, files in os.walk(source, followlinks=True):
        for name in files:
            if is_match(name, *patterns):
                file_path = os.path.abspath(os.path.join(path, name))
                replace_file_values_regex(file_path, data)


def replace_file_values(source, data):
    """
    replaces the values in given file with values available in given dict.

    the file must contain dict keys to be replaced by dict values for that keys.

    :param str source: file path to replace its values.
                       it must be an absolute path.

    :param dict[str, str] data: a dict containing all values that
                                must be replaced in given file.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    :raises PathNotExistedError: path not existed error.
    """

    path_utils.assert_exists(source)
    if data is not None and len(data) > 0:
        with open(source, 'r') as file:
            file_data = file.read()

        file_data = file_data.format(**data)

        with open(source, 'w') as file:
            file.write(file_data)


def replace_files_values(source, data, *patterns):
    """
    replaces the values in all files with values available in given dict.

    if patterns is given, only matching files will be included.

    :param str source: directory path to process its files.
                       the operation will also include all
                       files of all subdirectories.
                       it must be an absolute path.

    :param dict[str, str] data: a dict containing all values that
                                must be replaced in given file.

    :param str patterns: file name end patterns to be included
                         in replacement operation. for example
                         it could be `'.py', '.html'`.
                         all files will be included if not provided.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    :raises PathNotExistedError: path not existed error.
    :raises IsNotDirectoryError: is not directory error.
    """

    path_utils.assert_exists(source)
    if not os.path.isdir(source):
        raise IsNotDirectoryError('Provided path [{source}] is not a directory.'
                                  .format(source=source))

    for path, directories, files in os.walk(source, followlinks=True):
        for name in files:
            if is_match(name, *patterns):
                file_path = os.path.abspath(os.path.join(path, name))
                replace_file_values(file_path, data)


def is_match(source, *patterns):
    """
    gets a value indicating that given file name end, matches with any of given patterns.

    :param str source: source file path.

    :param str patterns: file name end pattern. for example it could
                         be `'.py', '.html'`. it will match all file
                         names if no pattern is provided.

    :rtype: bool
    """

    if len(patterns) <= 0:
        return True

    patterns = tuple(item.lower() for item in patterns)
    return source.lower().endswith(patterns)



# -*- coding: utf-8 -*-
"""
string normalizer handlers case module.
"""

from titlecase import titlecase

from pyrin.utilities.string.normalizer.decorators import string_normalizer
from pyrin.utilities.string.normalizer.enumerations import NormalizerEnum
from pyrin.utilities.string.normalizer.handlers.base import StringNormalizerBase


@string_normalizer()
class LowercaseNormalizer(StringNormalizerBase):
    """
    lowercase normalizer class.

    this normalizer makes the string lowercase.
    """

    def __init__(self, **options):
        """
        initializes an instance of LowercaseNormalizer.
        """

        super().__init__(NormalizerEnum.LOWERCASE, 210, **options)

    def _normalize(self, value, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :returns: normalized value.
        :rtype: str
        """

        return value.lower()


@string_normalizer()
class UppercaseNormalizer(StringNormalizerBase):
    """
    uppercase normalizer class.

    this normalizer makes the string uppercase.
    """

    def __init__(self, **options):
        """
        initializes an instance of UppercaseNormalizer.
        """

        super().__init__(NormalizerEnum.UPPERCASE, 230, **options)

    def _normalize(self, value, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :returns: normalized value.
        :rtype: str
        """

        return value.upper()


@string_normalizer()
class TitleCaseNormalizer(StringNormalizerBase):
    """
    title case normalizer class.

    this normalizer makes the string title case.
    meaning that first letter of each word will be capitalized.
    """

    def __init__(self, **options):
        """
        initializes an instance of TitleCaseNormalizer.
        """

        super().__init__(NormalizerEnum.TITLE_CASE, 220, **options)

    def _normalize(self, value, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :returns: normalized value.
        :rtype: str
        """

        return titlecase(value)



# -*- coding: utf-8 -*-
"""
utils configuration module.
"""

from configparser import ConfigParser

from pyrin.core.structs import DTO
from pyrin.utils.exceptions import InputNotCallableError, ConfigurationFileNotFoundError
from pyrin.utils.string import remove_line_break_escapes


def load(file_path, converter=eval, **options):
    """
    loads configurations from given config file into a dict.

    :param str file_path: file path to be loaded.

    :param callable converter: a callable to be used for converting
                               each value represented in config file.
                               the callable should accept a single argument.
                               defaults to `eval` if not provided.

    :keyword dict defaults: a dict containing values
                            needed for interpolation.
                            defaults to None if not provided.

    :raises InputNotCallableError: input not callable error.
    :raises ConfigurationFileNotFoundError: configuration file not found error.

    :rtype: dict
    """

    if not callable(converter):
        raise InputNotCallableError('Input parameter [{converter}] is not callable.'
                                    .format(converter=converter))

    defaults = options.get('defaults', None)
    parser = ConfigParser(defaults)
    if len(parser.read(file_path)) == 0:
        raise ConfigurationFileNotFoundError('Configuration file [{file}] not found.'
                                             .format(file=file_path))

    sections = DTO()
    for section in parser.sections():
        values = parser.items(section)
        dic_values = DTO()
        for single_value in values:
            value = converter(single_value[1])
            if isinstance(value, str):
                value = remove_line_break_escapes(value)
            dic_values[single_value[0]] = value

        sections[section] = dic_values

    return sections



# -*- coding: utf-8 -*-
"""
locale services module.
"""

from pyrin.application.services import get_component
from pyrin.globalization.locale import LocalePackage


def set_locale_selector(func):
    """
    sets the given function as locale selector.

    :param function func: function to be set as locale selector.

    :raises InvalidLocaleSelectorTypeError: invalid locale selector type error.

    :raises LocaleSelectorHasBeenAlreadySetError: locale selector has been
                                                  already set error.
    """

    get_component(LocalePackage.COMPONENT_NAME).set_locale_selector(func)


def set_timezone_selector(func):
    """
    sets the given function as timezone selector.

    :param function func: function to be set as timezone selector.

    :raises InvalidTimezoneSelectorTypeError: invalid timezone selector type error.

    :raises TimezoneSelectorHasBeenAlreadySetError: timezone selector has been
                                                    already set error.
    """

    get_component(LocalePackage.COMPONENT_NAME).set_timezone_selector(func)


def get_default_locale():
    """
    gets the default locale of application from `globalization` config store.

    :rtype: str
    """

    return get_component(LocalePackage.COMPONENT_NAME).get_default_locale()


def get_current_locale():
    """
    gets the current locale that should be used for current request.

    it never raises an error and returns the default locale if anything goes wrong.

    :rtype: str
    """

    return get_component(LocalePackage.COMPONENT_NAME).get_current_locale()


def get_current_timezone():
    """
    gets the current timezone that should be used for current request.

    it never raises an error and returns the default timezone if anything goes wrong.

    :rtype: str
    """

    return get_component(LocalePackage.COMPONENT_NAME).get_current_timezone()


def locale_exists(locale_name):
    """
    gets a value indicating that a locale with the given name exists.

    :param str locale_name: locale name to check for existence.

    :rtype: bool
    """

    return get_component(LocalePackage.COMPONENT_NAME).locale_exists(locale_name)



# -*- coding: utf-8 -*-
"""
schema component module.
"""

from pyrin.api.schema import SchemaPackage
from pyrin.api.schema.manager import SchemaManager
from pyrin.application.decorators import component
from pyrin.application.structs import Component


@component(SchemaPackage.COMPONENT_NAME)
class SchemaComponent(Component, SchemaManager):
    """
    schema component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
cli decorators module.
"""

from functools import update_wrapper

import pyrin.cli.services as cli_services

from pyrin.utils.custom_print import print_error


def cli(func):
    """
    decorator to specify a method that handles a cli command.

    the method name will be used as the handler name for that cli handler.
    the decorated method, will be get called if `help=True` is not in options,
    so you could do something in the decorated method body if required.
    but it's highly recommended that do not do any complex operation inside
    cli handler methods, instead implement the required operations in the manager
    of relevant package and just call the relevant service inside the cli method.

    :param function func: function.

    :raises CLIHandlerNotFoundError: cli handler not found error.
    :raises InvalidCLIDecoratedMethodError: invalid cli decorated method error.

    :returns: decorated function.
    :rtype: function
    """

    def decorator(*args, **kwargs):
        """
        decorates the given function and executes the relevant cli handler.

        :param object args: function arguments.
        :param object kwargs: function keyword arguments.

        :raises CLIHandlerNotFoundError: cli handler not found error.
        :raises InvalidCLIDecoratedMethodError: invalid cli decorated method error.

        :rtype: int
        """

        try:
            return cli_services.process_function(func, args, kwargs)
        except Exception as error:
            print_error(error, force=True)

    return update_wrapper(decorator, func)


def cli_invoke(func):
    """
    decorator to specify a method that will be invoked from cli.

    the decorated method will be get called by decorator and all errors
    will be printed to standard output instead of being raised.

    :param function func: function.

    :returns: decorated function.
    :rtype: function
    """

    def decorator(*args, **kwargs):
        """
        decorates the given function and executes it.

        :param object args: function arguments.
        :param object kwargs: function keyword arguments.
        """

        try:
            return cli_services.invoke_function(func, args, kwargs)
        except Exception as error:
            print_error(error, force=True)

    return update_wrapper(decorator, func)


def cli_group(name):
    """
    decorator to register a cli group.

    :param str name: cli group name.

    :raises CLIGroupNameIsRequiredError: cli group name is required error.
    :raises DuplicatedCLIGroupError: duplicated cli group error.
    :raises InvalidCLIGroupTypeError: invalid cli group type error.

    :returns: cli group class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available cli group.

        :param type cls: cli group class.

        :returns: cli group class.
        :rtype: type
        """

        instance = cls()
        cli_services.register_cli_group(name, instance)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
babel interface module.
"""

from pyrin.cli.base import CLIHandlerBase
from pyrin.cli.params import CLIParamBase


class BabelCLIParamBase(CLIParamBase):
    """
    babel cli param base class.

    all babel cli param classes must be subclassed from this.
    """
    pass


class BabelCLIHandlerBase(CLIHandlerBase):
    """
    babel cli handler base class.

    all babel cli handlers must be subclassed from this.
    """

    def __init__(self, name):
        """
        initializes an instance of BabelCLIHandlerBase.

        :param str name: the handler name that should be registered
                         with. this name must be the exact name that
                         this handler must emmit to cli.
        """

        super().__init__(name)

    def _get_common_cli_options(self):
        """
        gets the list of common cli options.

        :rtype: list
        """

        return ['pybabel']



# -*- coding: utf-8 -*-
"""
users manager module.
"""

from abc import abstractmethod

from pyrin.core.structs import Manager
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.security.users import UsersPackage


class UsersManager(Manager):
    """
    users manager class.
    this class is intended to provide some services needed in pyrin application.
    the top level application must extend this class considering business requirements.
    """

    package_class = UsersPackage

    @abstractmethod
    def is_active(self, user, **options):
        """
        gets a value indicating that given user is active.

        :param user: user to check its active status.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
security component module.
"""

from pyrin.application.decorators import component
from pyrin.security import SecurityPackage
from pyrin.security.manager import SecurityManager
from pyrin.application.structs import Component


@component(SecurityPackage.COMPONENT_NAME)
class SecurityComponent(Component, SecurityManager):
    """
    security component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
utils regex module.
"""

import re

from pyrin.core.globals import LIST_TYPES


def matches(pattern, string, flags=0):
    """
    gets all matches of given pattern in given string in a flat tuple.

    it also removes empty matches from the result.

    :param re.Pattern | str pattern: a pattern object or a regex string.
    :param str string: value to be searched for pattern.

    :param int flags: regex flags. note that flags could only be provided
                      if given pattern is a string.

    :rtype: tuple[str]
    """

    result = re.findall(pattern, string, flags=flags)
    flat_matches = []
    for item in result:
        if isinstance(item, LIST_TYPES):
            flat_matches.extend(item)
        else:
            flat_matches.append(item)

    return tuple(item for item in flat_matches if len(item) > 0)



# -*- coding: utf-8 -*-
"""
logging masking services module.
"""

from pyrin.application.services import get_component
from pyrin.logging.masking import LoggingMaskingPackage


def should_mask(key):
    """
    gets a value indicating that given key should be masked in a dict.

    :param str key: dict key to be checked.

    :rtype: bool
    """

    return get_component(LoggingMaskingPackage.COMPONENT_NAME).should_mask(key)


def mask(data, **options):
    """
    masks the given data if required.

    if data is not a dict, no changes will be made.

    :param dict | object data: data to be masked.

    :returns: a masked dict or original input data.
    :rtype: MaskedDict | object
    """

    return get_component(LoggingMaskingPackage.COMPONENT_NAME).mask(data, **options)



# -*- coding: utf-8 -*-
"""
caching local handlers permanent module.
"""

from pyrin.caching.local.containers.dict import OrderedDictContainer
from pyrin.caching.decorators import cache
from pyrin.caching.local.items.permanent import PermanentLocalCacheItem
from pyrin.caching.local.handlers.base import LocalCacheBase, ExtendedLocalCacheBase


@cache()
class PermanentLocalCache(LocalCacheBase):
    """
    permanent local cache class.

    this type of caches cache items permanently and they won't be expired.
    it does not consider method inputs, current user and component key for cache
    key generation.
    """

    cache_name = 'permanent'
    container_class = OrderedDictContainer
    cache_item_class = PermanentLocalCacheItem


@cache()
class ExtendedPermanentLocalCache(ExtendedLocalCacheBase):
    """
    extended permanent local cache class.

    this type of caches cache items permanently and they won't be expired.
    it does consider method inputs, current user and component key for cache key generation.
    """

    cache_name = 'extended.permanent'
    container_class = OrderedDictContainer
    cache_item_class = PermanentLocalCacheItem



# -*- coding: utf-8 -*-
"""
database migration manager module.
"""

from os import path

from sqlalchemy import MetaData

import pyrin.database.services as database_services
import pyrin.configuration.services as config_services
import pyrin.application.services as application_services
import pyrin.database.model.services as model_services

from pyrin.core.structs import DTO, Manager
from pyrin.database.migration import DatabaseMigrationPackage
from pyrin.database.migration.exceptions import EngineBindNameNotFoundError
from pyrin.utils.custom_print import print_warning, print_info


class DatabaseMigrationManager(Manager):
    """
    database migration manager class.
    """

    package_class = DatabaseMigrationPackage

    def __init__(self):
        """
        initializes an instance of DatabaseMigrationManager.
        """

        super().__init__()

        # a dictionary containing engine to tables map for all tables.
        # in the form of: {Engine engine: list[Table] tables}
        self._engine_to_table_map = DTO()

        # a dictionary containing bind name to metadata map. note that each metadata
        # only contains tables that should be created on relevant bounded engine.
        # the default engine is added with 'default' key.
        # in the form of: {str bind_name: MetaData metadata}
        self._bind_name_to_metadata_map = DTO()

    def create_all(self):
        """
        creates all entities on database engine.
        """

        for engine, tables in self._engine_to_table_map.items():
            if tables is not None and len(tables) > 0:
                model_services.get_metadata().create_all(engine, tables)

    def drop_all(self):
        """
        drops all entities on database engine.
        """

        for engine, tables in self._engine_to_table_map.items():
            if tables is not None and len(tables) > 0:
                model_services.get_metadata().drop_all(engine, tables)

    def _map_engine_to_table(self):
        """
        maps all engines to relevant tables.
        """

        all_tables = DTO(**model_services.get_tables())
        for entity, engine in database_services.get_entity_to_engine_map().items():
            if engine not in self._engine_to_table_map:
                self._engine_to_table_map[engine] = []

            self._engine_to_table_map[engine].append(all_tables.pop(entity.table_fullname))

        # all remaining tables are associated with default engine.
        if len(all_tables) > 0:
            self._engine_to_table_map[database_services.get_default_engine()] = \
                list(all_tables.values())

    def get_connection_urls(self):
        """
        gets all database connection urls for each engine.

        it gets the values from active section of each store.
        in case of sqlite usage, all urls will be made absolute.

        :returns: dict[str bind_name: str connection_url]
        :rtype: dict
        """

        prefix = database_services.get_configs_prefix()
        url_key = '{prefix}url'.format(prefix=prefix)
        connections = DTO()
        connections[database_services.get_default_database_name()] = \
            self._get_absolute_connection_url(config_services.get_active('database', url_key))

        for bind_name in self.get_bind_name_to_metadata_map():
            if bind_name != database_services.get_default_database_name():
                section_name = database_services.get_bind_config_section_name(bind_name)
                connections[bind_name] = self._get_absolute_connection_url(
                    config_services.get('database.binds', section_name, url_key))

        return connections

    def _map_bind_to_metadata(self):
        """
        maps bind names of different engines to a metadata representing the related tables.
        """

        for engine, tables in self._engine_to_table_map.items():
            bounded_engines = database_services.get_bounded_engines()
            bind_name_map = None
            for bind_name, bounded_engine in bounded_engines.items():
                if engine == bounded_engine:
                    bind_name_map = bind_name
                    break

            default_engine = database_services.get_default_engine()
            if bind_name_map is None and engine == default_engine:
                bind_name_map = database_services.get_default_database_name()
            elif bind_name_map is None and engine != default_engine:
                raise EngineBindNameNotFoundError('Could not find any bind name '
                                                  'for database engine [{name}].'
                                                  .format(name=str(engine)))

            self._bind_name_to_metadata_map[bind_name_map] = self._add_to_metadata(tables)

    def get_bind_name_to_metadata_map(self):
        """
        gets bind name to metadata map.

        :returns: dict[str bind_name: MetaData metadata]
        :rtype: dict
        """

        return self._bind_name_to_metadata_map

    def _add_to_metadata(self, tables):
        """
        adds given tables into a metadata and returns it.

        :param list[Tables] tables: tables to be added into metadata.

        :rtype: MetaData
        """

        metadata = MetaData()
        for table in tables:
            table.tometadata(metadata)

        return metadata

    def configure_migration_data(self):
        """
        configures the required data for any migration.
        """

        self._map_engine_to_table()
        self._map_bind_to_metadata()

    def _get_absolute_connection_url(self, connection_url):
        """
        gets absolute path of given connection string if required.

        it is only required when sqlite is in use.

        :param str connection_url: connection url.

        :rtype: str
        """

        if connection_url is None or 'sqlite' not in connection_url.lower() or \
                connection_url.lower().startswith('sqlite:////') or \
                'sqlite:///:memory:' in connection_url.lower() or \
                connection_url == 'sqlite:///':
            return connection_url

        root_path = application_services.get_working_directory()
        url = connection_url.replace('sqlite:///', '')
        full_path = path.join(root_path, url)
        absolute_path = path.abspath(full_path)

        return 'sqlite:///{absolute_path}'.format(absolute_path=absolute_path)

    def rebuild_schema(self):
        """
        rebuilds database schema based on database and environment config store values.

        this will occur only if application is not in scripting mode.
        """

        if application_services.is_scripting_mode() is False:
            if config_services.get_active('database', 'drop_on_startup') is True:
                environment = config_services.get_active('environment', 'env')
                debug = config_services.get_active('environment', 'debug')
                unit_testing = config_services.get_active('environment', 'unit_testing')

                if (environment == 'development' and debug is True) or \
                        (environment == 'testing' and unit_testing is True):
                    print_warning('Dropping all models...')
                    self.drop_all()

            if config_services.get_active('database', 'create_on_startup') is True:
                print_info('Creating all models...')
                self.create_all()



# -*- coding: utf-8 -*-
"""
deserializer handlers uuid module.
"""

import pyrin.utils.unique_id as uuid_utils

from pyrin.core.globals import NULL
from pyrin.utils.unique_id import UUID_REGEX
from pyrin.converters.deserializer.decorators import deserializer
from pyrin.converters.deserializer.handlers.base import StringPatternDeserializerBase


@deserializer()
class UUIDDeserializer(StringPatternDeserializerBase):
    """
    uuid deserializer class.
    """

    def __init__(self, **options):
        """
        creates an instance of UUIDDeserializer.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: a list of custom accepted
                                                                  formats and their min and
                                                                  max length for uuid
                                                                  deserialization.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        returns `NULL` object if deserialization fails.

        :param str value: value to be deserialized.

        :keyword Pattern matching_pattern: the pattern that has matched the value.

        :rtype: uuid.UUID
        """

        result = uuid_utils.try_get_uuid(value)
        if result is None:
            return NULL

        return result

    @property
    def default_formats(self):
        """
        gets default accepted formats that this deserializer could deserialize value from.

        :returns: list[tuple[Pattern format, int min_length, int max_length]]
        :rtype: list[tuple[Pattern, int, int]]
        """

        return [(UUID_REGEX, 36, 36)]



# -*- coding: utf-8 -*-
"""
response enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class ResponseHeaderEnum(CoreEnum):
    """
    response header enum.
    """

    # Accept-Patch: text/example;charset=utf-8
    # specifies which patch document formats this server supports.
    ACCEPT_PATCH = 'Accept-Patch'

    # Accept-Ranges: bytes
    # what partial content range types this server supports via byte serving.
    ACCEPT_RANGES = 'Accept-Ranges'

    # Age: 12
    # the age the object has been in a proxy cache in seconds.
    AGE = 'Age'

    # Allow: GET, HEAD
    # valid methods for a specified resource. to be used for a 405 Method not allowed.
    ALLOW = 'Allow'

    # Alt-Svc: http/1.1= "http2.example.com:8001"; ma=7200
    # a server uses “Alt-Svc” header (meaning alternative services) to indicate that its
    # resources can also be accessed at a different network location (host or port) or using
    # a different protocol. when using HTTP/2, servers should instead send an ALTSVC frame.
    ALT_SVC = 'Alt-Svc'

    # Cache-Control: max-age=3600 Cache-Control: no-cache, no-store, max-age=0, must-revalidate
    # if no-cache is used, the 'Cache-Control' header can tell the browser to never use a cached
    # version of a resource without first checking the ETag value.
    # 'max-age' is measured in seconds. the more restrictive 'no-store' option tells the browser
    # (and all the intermediary network devices) not even store the resource in its cache.
    CACHE_CONTROL = 'Cache-Control'

    # Connection: close
    # Control options for the current connection and list of hop-by-hop response fields.
    # deprecated in HTTP/2
    CONNECTION = 'Connection'

    # Content-Disposition: attachment; filename="file.txt"
    # an opportunity to raise a 'File Download' dialogue box for a known MIME type with
    # binary format or suggest a filename for dynamic content. quotes are necessary with
    # special characters.
    CONTENT_DISPOSITION = 'Content-Disposition'

    # Content-Encoding: gzip
    # the type of encoding used on the data.
    CONTENT_ENCODING = 'Content-Encoding'

    # Content-Language: en
    # the natural language or languages of the intended audience for the enclosed content.
    CONTENT_LANGUAGE = 'Content-Language'

    # Content-Length: 348
    # the length of the response body expressed in 8-bit bytes.
    CONTENT_LENGTH = 'Content-Length'

    # Content-Location: /index.htm
    # an alternate location for the returned data.
    CONTENT_LOCATION = 'Content-Location'

    # Content-Range: bytes 21010-47021/47022
    # where in a full body message this partial message belongs.
    CONTENT_RANGE = 'Content-Range'

    # Content-Type: text/html; charset=utf-8
    # the mime type of this content.
    CONTENT_TYPE = 'Content-Type'

    # Date: Tue, 15 Nov 1994 08:12:31 GMT
    # the date and time that the message was sent.
    # (in 'HTTP-date' format as defined by RFC 7231).
    DATE = 'Date'

    # Delta-Base: "abc"
    # specifies the delta-encoding entity tag of the response.
    DELTA_BASE = 'Delta-Base'

    # ETag: "737060cd8c284d8a[...]"
    # an identifier for a specific version of a resource, often a message digest.
    ETAG = 'ETag'

    # Expires: Sat, 01 Dec 2018 16:00:00 GMT
    # gives the date/time after which the response is considered stale.
    # (in 'HTTP-date' format as defined by RFC 7231).
    EXPIRES = 'Expires'

    # IM: feed
    # instance-manipulations applied to the response.
    IM = 'IM'

    # Last-Modified: Mon, 15 Nov 2017 12:00:00 GMT
    # the last modified date for the requested object.
    # (in 'HTTP-date' format as defined by RFC 7231).
    LAST_MODIFIED = 'Last-Modified'

    # Link: </feed>; rel="alternate"
    # used to express a typed relationship with another resource, where the relation
    # type is defined by RFC 5988.
    LINK = 'Link'

    # Location: /pub/WWW/People.html
    # used in redirection, or when a new resource has been created.
    LOCATION = 'Location'

    # Pragma: no-cache
    # implementation-specific fields that may have various effects anywhere along
    # the request-response chain.
    PRAGMA = 'Pragma'

    # Proxy-Authenticate: Basic
    # request authentication to access the proxy.
    PROXY_AUTHENTICATE = 'Proxy-Authenticate'

    # HTTP public key pinning, announces hash of website’s authentic TLS certificate.
    PUBLIC_KEY_PINS = 'Public-Key-Pins'

    # Retry-After: 120 Retry-After: Fri, 07 Nov 2014 23:59:59 GMT
    # if an entity is temporarily unavailable, this instructs the client to try again later.
    # value could be a specified period of time (in seconds) or an 'HTTP-date'.
    RETRY_AFTER = 'Retry-After'

    # Server: Apache/2.4.1 (Unix)
    # a name for the server.
    SERVER = 'Server'

    # Set-Cookie: UserID=JohnDoe; Max-Age=3600; Version=1
    # an HTTP cookie.
    SET_COOKIE = 'Set-Cookie'

    # Strict-Transport-Security: max-age=16070400; includeSubDomains
    # an HSTS policy informing the HTTP client how long to cache the HTTPS only policy
    # and whether this applies to subdomains.
    STRICT_TRANSPORT_SECURITY = 'Strict-Transport-Security'

    # Trailer: Max-Forwards
    # the trailer general field value indicates that the given set of header fields is
    # present in the trailer of a message encoded with chunked transfer coding.
    TRAILER = 'Trailer'

    # Transfer-Encoding: chunked
    # the form of encoding used to safely transfer the entity to the user.
    # currently defined methods are: chunked, compress, deflate, gzip, identity.
    # deprecated in HTTP/2.
    TRANSFER_ENCODING = 'Transfer-Encoding'

    # Tk: ?
    # tracking status header, value suggested to be sent in response to a DNT(do-not-track),
    # possible values: '!' — under construction. '?' — dynamic. 'G' — gateway to multiple parties.
    # 'N' — not tracking. 'T' — tracking. 'C' — tracking with consent. 'P' — tracking only if
    # consented. 'D' — disregarding DNT. 'U' — updated.
    TK = 'Tk'

    # Upgrade: h2c, HTTPS/1.3, IRC/6.9, RTA/x11, websocket
    # ask the client to upgrade to another protocol.
    # deprecated in HTTP/2.
    UPGRADE = 'Upgrade'

    # Vary: Accept-Language Vary: *
    # tells downstream proxies how to match future request headers to decide whether the
    # cached response can be used rather than requesting a fresh one from the origin server.
    VARY = 'Vary'

    # Via: 1.0 fred, 1.1 example.com (Apache/1.1)
    # informs the client of proxies through which the response was sent.
    VIA = 'Via'

    # Warning: 199 Miscellaneous warning
    # a general warning about possible problems with the entity body.
    WARNING = 'Warning'

    # WWW-Authenticate: Basic
    # indicates the authentication scheme that should be used to access the requested entity.
    WWW_AUTHENTICATE = 'WWW-Authenticate'


class NonStandardResponseHeaderEnum(CoreEnum):
    """
    non standard response header enum.
    """

    # helps to protect against XSS attacks.
    CONTENT_SECURITY_POLICY = 'Content-Security-Policy'

    # Refresh: 10;http://www.example.org/
    # redirect to a URL after an arbitrary delay expressed in seconds.
    REFRESH = 'Refresh'

    # X-Powered-By: Brain/0.6b
    # can be used by servers to send their name and version.
    X_POWERED_BY = 'X-Powered-By'

    # allows the server to pass a request id that clients can send
    # back to let the server correlate the request.
    X_REQUEST_ID = 'X-Request-ID'

    # sets which version of Internet Explorer compatibility layer should be used.
    # only used if you need to support IE8 or IE9.
    X_UA_COMPATIBLE = 'X-UA-Compatible'

    # now replaced by the 'Content-Security-Policy' header, used in older browsers
    # to stop pages load when an XSS attack is detected.
    X_XSS_PROTECTION = 'X-XSS-Protection'



# -*- coding: utf-8 -*-
"""
router handlers public module.
"""

from pyrin.api.router.handlers.base import RouteBase, TemporaryRouteBase


class PublicRoute(RouteBase):
    """
    public route class.

    this class should be used to manage application public api
    routes that do not require authentication.
    """
    pass


class PublicTemporaryRoute(PublicRoute, TemporaryRouteBase):
    """
    public temporary route class.

    this class should be used to manage application public api
    routes that do not require authentication and must be unregistered
    after a specified number of requests or a duration of time.
    this is useful for a couple of different things but mostly for system
    health checking after each new deployment on production.
    """
    pass



# -*- coding: utf-8 -*-
"""
response manager module.
"""

from werkzeug.datastructures import Headers
from flask import Response

import pyrin.utils.headers as header_utils

from pyrin.core.enumerations import ServerErrorResponseCodeEnum
from pyrin.core.globals import ROW_RESULT
from pyrin.processor.response import ResponsePackage
from pyrin.settings.static import DEFAULT_STATUS_CODE
from pyrin.core.structs import Manager


class ResponseManager(Manager):
    """
    response manager class.
    """

    package_class = ResponsePackage

    def _render_body(self, **options):
        """
        renders the given keyword arguments as response body.

        this method renders the given arguments as dict for json serialization.
        it simply returns the input dict as result.
        but if you want to render body to other formats, for example html,
        you could override this method and render the body as you need.

        :param object options: all keyword arguments that
                               must be rendered as response body.

        :rtype: dict
        """

        return options

    def make_response(self, **options):
        """
        makes a response from given inputs.

        note that every other keyword argument that is present
        in given input, will be passed to generated response.

        :keyword int code: response code.
                           defaults to `DEFAULT_STATUS_CODE`
                           code, if not provided.

        :keyword dict | Headers headers: headers to add into response.

        :returns: tuple[dict | object, int, CoreHeaders]
        :rtype: tuple
        """

        code = options.get('code', None)
        if code is None:
            code = DEFAULT_STATUS_CODE
            options.update(code=code)

        headers = options.pop('headers', None)
        headers = header_utils.convert_headers(headers)
        body = self._render_body(**options)
        return self.pack_response(body, code, headers)

    def make_error_response(self, message, **options):
        """
        makes an error response from given inputs.

        :param str message: error message.

        :keyword int code: error code.
                           defaults to `INTERNAL_SERVER_ERROR`
                           code, if not provided.

        :keyword dict | Headers headers: headers to add into response.

        :returns: tuple[dict | object, int, CoreHeaders]
        :rtype: tuple
        """

        code = options.get('code', None)
        if code is None:
            code = ServerErrorResponseCodeEnum.INTERNAL_SERVER_ERROR
            options.update(code=code)

        data = options.get('data', None)
        if data is None:
            data = {}

        options.update(message=message, data=data)
        return self.make_response(**options)

    def make_exception_response(self, exception, **options):
        """
        makes an error response from given exception.

        if the exception does not have code, defaults
        to `INTERNAL_SERVER_ERROR` code.

        :param Exception exception: exception instance.

        :keyword int code: error code.
                           tries to get it from exception itself, if not
                           provided. defaults to `INTERNAL_SERVER_ERROR`
                           code if not available in exception.

        :keyword dict | Headers headers: headers to add into response.

        :returns: tuple[dict | object, int, CoreHeaders]
        :rtype: tuple
        """

        message = getattr(exception, 'description', str(exception))
        code = options.get('code', None)
        if code is None:
            code = getattr(exception, 'code',
                           ServerErrorResponseCodeEnum.INTERNAL_SERVER_ERROR)
            options.update(code=code)

        return self.make_error_response(message, **options)

    def unpack_response(self, response, **options):
        """
        unpacks the response object into a tuple of three parts.

        in the form of `(body, status_code, headers)`. if any of these
        parts are not present in provided response, it returns None for
        that specific part.
        note that if a view function returns a `Response` object, the status
        code will be fetched from that object if available. but if a view function
        returns (Response response, int status_code) then the stand-alone status
        code will override the status code of `Response` object.

        :param tuple | object response: response object to be unpacked.

        :returns: tuple[object body, int status_code, CoreHeaders headers]
        :rtype: tuple[object, int, CoreHeaders]
        """

        body = None
        status_code = None
        headers = None
        if isinstance(response, tuple) and \
                not isinstance(response, ROW_RESULT) and len(response) in (2, 3):
            length = len(response)
            if length == 3:
                body, status_code, headers = response
            else:
                if isinstance(response[1], (Headers, dict, tuple, list)):
                    body, headers = response
                else:
                    body, status_code = response
        else:
            body = response

        if status_code is None and isinstance(body, Response):
            if body.status_code not in (None, 0):
                status_code = body.status_code

        headers = header_utils.convert_headers(headers)
        return body, status_code, headers

    def pack_response(self, body, status_code, headers, **options):
        """
        packs the response using given values.

        it returns a tuple if status code or headers are
        not None, otherwise it just returns the body.

        :param object | CoreResponse body: body of response.
        :param int status_code: status code of response.
        :param dict | Headers headers: all response headers.

        :returns: tuple[object body, int status_code, CoreHeaders headers] | object
        :rtype: tuple[object, int, CoreHeaders] | object
        """

        headers = header_utils.convert_headers(headers)

        if status_code is not None and headers is not None:
            return body, status_code, headers

        if status_code is not None:
            return body, status_code

        if headers is not None:
            return body, headers

        return body

    def get_body(self, response, **options):
        """
        gets the first part of a response tuple.

        it could be the body data or a response object itself.
        if response is not a tuple, the return value is the response itself.

        :param tuple | CoreResponse | object response: the response object
                                                       or instance or tuple.

        :rtype: CoreResponse | object
        """

        body, status, headers = self.unpack_response(response, **options)
        return body



# -*- coding: utf-8 -*-
"""
rs256 token handler module.
"""

import pyrin.configuration.services as config_services

from pyrin.security.token.decorators import token
from pyrin.security.token.handlers.base import RSTokenBase


@token()
class RS256Token(RSTokenBase):
    """
    rs256 token class.
    """

    def __init__(self, **options):
        """
        initializes an instance of RS256Token.
        """

        super().__init__(**options)

    def _get_encoding_key(self, **options):
        """
        gets the signing key for encoding.

        :rtype: str
        """

        return self._private_key

    def _get_decoding_key(self, **options):
        """
        gets the signing key for decoding.

        :rtype: str
        """

        return self._public_key

    def _get_algorithm(self, **options):
        """
        gets the algorithm for signing the token.

        :rtype: str
        """

        return 'RS256'

    def get_kid(self):
        """
        gets kid value to be used in token header for this handler.
        it must be unique for each handler.

        :rtype: str
        """

        return '8441c8a0-8ec6-4987-ab80-be71c3bef90c'

    def generate_key(self, **options):
        """
        generates a valid public/private key for this handler and returns it.

        :returns: tuple[str public_key, str private_key]
        :rtype: tuple[str, str]
        """

        return super().generate_key(length=2048)

    def _load_keys(self, **options):
        """
        loads public/private keys into this class's relevant attributes.
        """

        self._public_key = config_services.get('security', 'token', 'rs256_public_key')
        self._private_key = config_services.get('security', 'token', 'rs256_private_key')



# -*- coding: utf-8 -*-
"""
sentry package.
"""

from pyrin.packaging.base import Package


class SentryPackage(Package):
    """
    sentry package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'logging.sentry.component'
    CONFIG_STORE_NAMES = ['sentry']
    DEPENDS = ['pyrin.configuration',
               'pyrin.api']



# -*- coding: utf-8 -*-
"""
alembic handlers revision module.
"""

from pyrin.database.migration.alembic.enumerations import AlembicCLIHandlersEnum
from pyrin.database.migration.alembic.interface import AlembicCLIHandlerBase
from pyrin.database.migration.alembic.decorators import alembic_cli_handler
from pyrin.database.migration.alembic.handlers.params import MessageParam, \
    AutoGenerateParam, SQLParam, HeadParam, SpliceParam, BranchLabelParam, \
    VersionPathParam, RevisionIDParam, DependsOnParam


@alembic_cli_handler()
class RevisionCLIHandler(AlembicCLIHandlerBase):
    """
    revision cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of RevisionCLIHandler.
        """

        super().__init__(AlembicCLIHandlersEnum.REVISION)

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        :param list[CLIParamBase] params: list of all params.
        """

        params.extend([MessageParam(), BranchLabelParam(),
                       RevisionIDParam(), SpliceParam(),
                       AutoGenerateParam(default=True),
                       HeadParam(), DependsOnParam(),
                       VersionPathParam(), SQLParam()])

        return super()._inject_params(params)



# -*- coding: utf-8 -*-
"""
session services module.
"""

from pyrin.application.services import get_component
from pyrin.security.session import SessionPackage


def get_current_user():
    """
    gets current user.
    """

    return get_component(SessionPackage.COMPONENT_NAME).get_current_user()


def set_current_user(user):
    """
    sets current user.

    :param user: user object.

    :raises InvalidUserError: invalid user error.
    :raises CouldNotOverwriteCurrentUserError: could not overwrite current user error.
    """

    return get_component(SessionPackage.COMPONENT_NAME).set_current_user(user)


def get_current_request():
    """
    gets current request object.

    :raises RuntimeError: runtime error.

    :rtype: CoreRequest
    """

    return get_component(SessionPackage.COMPONENT_NAME).get_current_request()


def get_current_request_id():
    """
    gets current request id.

    this method is implemented to overcome the hash problem of requests.
    `CoreRequest` objects are hashable themselves, but when they used as a dict key
    some hash collisions will occur. so we have to expose the exact request id to
    be able to use it as a dict key in places such as database scoped sessions.

    :raises RuntimeError: runtime error.

    :rtype: uuid.UUID
    """

    return get_component(SessionPackage.COMPONENT_NAME).get_current_request_id()


def add_request_context(key, value, **options):
    """
    adds the given key/value pair into current request context.

    :param str key: key name to be added.
    :param object value: value to be added.

    :keyword bool replace: specifies that if a key with the same name
                           is already present, replace it. otherwise
                           raise an error. defaults to False if not provided.

    :raises InvalidRequestContextKeyNameError: invalid request context key name error.
    :raises RequestContextKeyIsAlreadyPresentError: request context key is
                                                    already present error.
    """

    get_component(SessionPackage.COMPONENT_NAME).add_request_context(key, value, **options)


def get_request_context(key, default=None):
    """
    gets the value for given key from current request context.

    it gets the default value if key is not present in the request context.

    :param str key: key name to get its value.
    :param object default: a value to be returned if the provided
                           key is not present in request context.

    :returns: object
    """

    return get_component(SessionPackage.COMPONENT_NAME).get_request_context(key, default)


def remove_request_context(key):
    """
    removes the specified key from current request context if available.

    :param str key: key name to be removed from request context.
    """

    get_component(SessionPackage.COMPONENT_NAME).remove_request_context(key)


def is_fresh():
    """
    gets a value indicating that current request has a fresh authentication.

    fresh authentication means an authentication which is done by providing
    user credentials to server.

    :rtype: bool
    """

    return get_component(SessionPackage.COMPONENT_NAME).is_fresh()


def get_current_token_payload():
    """
    gets current request's token payload.

    :rtype: dict
    """

    return get_component(SessionPackage.COMPONENT_NAME).get_current_token_payload()


def get_current_token_header():
    """
    gets current request's token header.

    :rtype: dict
    """

    return get_component(SessionPackage.COMPONENT_NAME).get_current_token_header()


def set_component_custom_key(value):
    """
    sets the component custom key.

    :param object value: component custom key value.

    :raises InvalidComponentCustomKeyError: invalid component custom key error.
    """

    return get_component(SessionPackage.COMPONENT_NAME).set_component_custom_key(value)


def get_component_custom_key():
    """
    gets component custom key.

    :rtype: object
    """

    return get_component(SessionPackage.COMPONENT_NAME).get_component_custom_key()


def get_safe_component_custom_key():
    """
    gets component custom key in a safe manner.

    meaning that if the request does not exist in current context, it will
    return a None object instead of raising an error.

    :rtype: object
    """

    return get_component(SessionPackage.COMPONENT_NAME).get_safe_component_custom_key()


def get_safe_current_request():
    """
    gets current request object in a safe manner.

    meaning that if the request does not exist in current context, it will
    return a None object instead of raising an error.

    :rtype: CoreRequest
    """

    return get_component(SessionPackage.COMPONENT_NAME).get_safe_current_request()


def get_safe_current_user():
    """
    gets current user in a safe manner.

    meaning that if the request does not exist in current context, it will
    return a None object instead of raising an error.
    """

    return get_component(SessionPackage.COMPONENT_NAME).get_safe_current_user()


def get_safe_cacheable_current_user():
    """
    gets cacheable current user in a safe manner.

    meaning that if the request does not exist in current context, it will
    return a None object instead of raising an error.
    """

    return get_component(SessionPackage.COMPONENT_NAME).get_safe_cacheable_current_user()


def is_request_context_available():
    """
    gets a value indicating that request context is available for usage.

    :rtype: bool
    """

    return get_component(SessionPackage.COMPONENT_NAME).is_request_context_available()



# -*- coding: utf-8 -*-
"""
validator handlers boolean module.
"""

from pyrin.core.globals import _
from pyrin.validator.handlers.base import ValidatorBase
from pyrin.validator.handlers.exceptions import ValueIsNotBooleanError


class BooleanValidator(ValidatorBase):
    """
    boolean validator class.
    """

    invalid_type_error = ValueIsNotBooleanError
    invalid_type_message = _('The provided value for [{param_name}] '
                             'must be of boolean type.')

    def __init__(self, domain, field, **options):
        """
        initializes an instance of BooleanValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param InstrumentedAttribute | str field: validator field name. it could be a
                                                  string or a column. each validator will
                                                  be registered with its field name in
                                                  corresponding domain. to enable automatic
                                                  validations, the provided field name must
                                                  be the exact name of the parameter which
                                                  this validator will validate. if you pass
                                                  a column attribute, some constraints
                                                  such as `nullable`, `min_length`, `max_length`,
                                                  `min_value`, `max_value`, `allow_blank`,
                                                  `allow_whitespace`, `check_in` and
                                                  `check_not_in` could be extracted
                                                  automatically from that column if not provided
                                                  in inputs.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :keyword str name: a custom name for this validator.
                           if provided, the name of `field` will be ignored.

        :keyword bool for_find: specifies that this validator must only
                                be used on validation for find.
                                defaults to False if not provided.

        :raises ValidatorFieldIsRequiredError: validator field is required error.
        :raises ValidatorNameIsRequiredError: validator name is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises InvalidNotAcceptedTypeError: invalid not accepted type error.
        :raises ValidatorFixerMustBeCallable: validator fixer must be callable.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        """

        options.update(accepted_type=bool)
        super().__init__(domain, field, **options)
